@@ -619,11 +619,21 @@ get_data2_autocor.brmsformula <- function(x, ...) {
   attr(attr(x$formula, "autocor"), "data2")
 }
 
+#' @export
+get_data2_autocor.brmsterms <- function(x, ...) {
+  attr(attr(x$formula, "autocor"), "data2")
+}
+
 #' @export
 get_data2_autocor.mvbrmsformula <- function(x, ...) {
   ulapply(x$forms, get_data2_autocor, recursive = FALSE)
 }
 
+#' @export
+get_data2_autocor.mvbrmsterms <- function(x, ...) {
+  ulapply(x$terms, get_data2_autocor, recursive = FALSE)
+}
+
 #' @export
 print.cor_brms_formula <- function(x, ...) {
   y <- x
