@@ -24,9 +24,9 @@ parse_model <- function(model, backend, ...) {
 .parse_model_cmdstanr <- function(model, silent = 1, ...) {
   require_package("cmdstanr")
   temp_file <- cmdstanr::write_stan_file(model)
-  if (cmdstanr::cmdstan_version() >= "2.29.0") {
-    .canonicalize_stan_model(temp_file, overwrite_file = TRUE)
-  }
+  # if (cmdstanr::cmdstan_version() >= "2.29.0") {
+  #   .canonicalize_stan_model(temp_file, overwrite_file = TRUE)
+  # }
   out <- eval_silent(
     cmdstanr::cmdstan_model(temp_file, compile = FALSE, ...),
     type = "message", try = TRUE, silent = silent
@@ -70,7 +70,7 @@ compile_model <- function(model, backend, ...) {
   if (silent < 2) {
     message("Compiling Stan program...")
   }
-  if (use_threading(threads)) {
+  if (use_threading(threads, force = TRUE)) {
     if (utils::packageVersion("rstan") >= "2.26") {
       threads_per_chain_def <- rstan::rstan_options("threads_per_chain")
       on.exit(rstan::rstan_options(threads_per_chain = threads_per_chain_def))
@@ -97,10 +97,10 @@ compile_model <- function(model, backend, ...) {
   require_package("cmdstanr")
   args <- list(...)
   args$stan_file <- cmdstanr::write_stan_file(model)
-  if (cmdstanr::cmdstan_version() >= "2.29.0") {
-    .canonicalize_stan_model(args$stan_file, overwrite_file = TRUE)
-  }
-  if (use_threading(threads)) {
+  # if (cmdstanr::cmdstan_version() >= "2.29.0") {
+  #   .canonicalize_stan_model(args$stan_file, overwrite_file = TRUE)
+  # }
+  if (use_threading(threads, force = TRUE)) {
     args$cpp_options$stan_threads <- TRUE
   }
   if (use_opencl(opencl)) {
@@ -147,7 +147,7 @@ fit_model <- function(model, backend, ...) {
                              seed, control, silent, future, ...) {
 
   # some input checks and housekeeping
-  if (use_threading(threads)) {
+  if (use_threading(threads, force = TRUE)) {
     if (utils::packageVersion("rstan") >= "2.26") {
       threads_per_chain_def <- rstan::rstan_options("threads_per_chain")
       on.exit(rstan::rstan_options(threads_per_chain = threads_per_chain_def))
@@ -166,6 +166,7 @@ fit_model <- function(model, backend, ...) {
   } else if (is.character(init) && !init %in% c("random", "0")) {
     init <- get(init, mode = "function", envir = parent.frame())
   }
+  future <- future && algorithm %in% "sampling"
   args <- nlist(
     object = model, data = sdata, iter, seed,
     init = init, pars = exclude, include = FALSE
@@ -187,7 +188,7 @@ fit_model <- function(model, backend, ...) {
         warning2("Argument 'cores' is ignored when using 'future'.")
       }
       args$chains <- 1L
-      futures <- fits <- vector("list", chains)
+      out <- futures <- vector("list", chains)
       for (i in seq_len(chains)) {
         args$chain_id <- i
         if (is.list(init)) {
@@ -200,10 +201,10 @@ fit_model <- function(model, backend, ...) {
         )
       }
       for (i in seq_len(chains)) {
-        fits[[i]] <- future::value(futures[[i]])
+        out[[i]] <- future::value(futures[[i]])
       }
-      out <- rstan::sflist2stanfit(fits)
-      rm(futures, fits)
+      out <- rstan::sflist2stanfit(out)
+      rm(futures)
     } else {
       c(args) <- nlist(chains, cores)
       out <- do_call(rstan::sampling, args)
@@ -215,6 +216,7 @@ fit_model <- function(model, backend, ...) {
   } else {
     stop2("Algorithm '", algorithm, "' is not supported.")
   }
+  # TODO: add support for pathfinder and laplace
   out <- repair_stanfit(out)
   out
 }
@@ -238,17 +240,8 @@ fit_model <- function(model, backend, ...) {
   } else if (is_equal(init, "0")) {
     init <- 0
   }
-  if (future) {
-    stop2("Argument 'future' is not supported by backend 'cmdstanr'.")
-  }
+  future <- future && algorithm %in% "sampling"
   args <- nlist(data = sdata, seed, init)
-  if (use_threading(threads)) {
-    if (algorithm %in% c("sampling", "fixed_param")) {
-      args$threads_per_chain <- threads$threads
-    } else if (algorithm %in% c("fullrank", "meanfield")) {
-      args$threads <- threads$threads
-    }
-  }
   if (use_opencl(opencl)) {
     args$opencl_ids <- opencl$ids
   }
@@ -272,37 +265,79 @@ fit_model <- function(model, backend, ...) {
   if (silent < 2) {
     message("Start sampling")
   }
+  use_threading <- use_threading(threads, force = TRUE)
   if (algorithm %in% c("sampling", "fixed_param")) {
     c(args) <- nlist(
       iter_sampling = iter - warmup,
       iter_warmup = warmup,
       chains, thin,
       parallel_chains = cores,
-      show_messages = !silent,
+      show_messages = silent < 2,
+      show_exceptions = silent == 0,
       fixed_param = algorithm == "fixed_param"
     )
-    out <- do_call(model$sample, args)
+    if (use_threading) {
+      args$threads_per_chain <- threads$threads
+    }
+    if (future) {
+      if (cores > 1L) {
+        warning2("Argument 'cores' is ignored when using 'future'.")
+      }
+      args$chains <- 1L
+      out <- futures <- vector("list", chains)
+      for (i in seq_len(chains)) {
+        args$chain_ids <- i
+        if (is.list(init)) {
+          args$init <- init[i]
+        }
+        futures[[i]] <- future::future(
+          brms::do_call(model$sample, args),
+          packages = "cmdstanr",
+          seed = TRUE
+        )
+      }
+      for (i in seq_len(chains)) {
+        out[[i]] <- future::value(futures[[i]])
+      }
+      rm(futures)
+    } else {
+      out <- do_call(model$sample, args)
+    }
   } else if (algorithm %in% c("fullrank", "meanfield")) {
-    # vb does not support parallel execution
     c(args) <- nlist(iter, algorithm)
+    if (use_threading) {
+      args$threads <- threads$threads
+    }
     out <- do_call(model$variational, args)
+  } else if (algorithm %in% c("pathfinder")) {
+    if (use_threading) {
+      args$num_threads <- threads$threads
+    }
+    out <- do_call(model$pathfinder, args)
+  } else if (algorithm %in% c("laplace")) {
+    if (use_threading) {
+      args$threads <- threads$threads
+    }
+    out <- do_call(model$laplace, args)
   } else {
     stop2("Algorithm '", algorithm, "' is not supported.")
   }
-  # not all metadata is not stored by read_csv_as_stanfit
-  metadata <- cmdstanr::read_cmdstan_csv(
-    out$output_files(), variables = "", sampler_diagnostics = ""
+
+  if (future) {
+    # 'out' is a list of fitted models
+    output_files <- ulapply(out, function(x) x$output_files())
+    stan_variables <- out[[1]]$metadata()$stan_variables
+  } else {
+    # 'out' is a single fitted model
+    output_files <- out$output_files()
+    stan_variables <- out$metadata()$stan_variables
+  }
+
+  out <- read_csv_as_stanfit(
+    output_files, variables = stan_variables,
+    model = model, exclude = exclude, algorithm = algorithm
   )
-  # ensure that only relevant variables are read from CSV
-  variables <- repair_variable_names(metadata$metadata$variables)
-  variables <- unique(sub("\\[.+", "", variables))
-  variables <- setdiff(variables, exclude)
-  # transform into stanfit object for consistent output structure
-  out <- read_csv_as_stanfit(out$output_files(), variables = variables)
-  out <- repair_stanfit(out)
-  # allow updating the model without recompilation
-  attributes(out)$CmdStanModel <- model
-  attributes(out)$metadata <- metadata
+
   if (empty_model) {
     # allow correct updating of an 'empty' model
     out@sim <- list()
@@ -423,7 +458,7 @@ backend_choices <- function() {
 
 # supported Stan algorithms
 algorithm_choices <- function() {
-  c("sampling", "meanfield", "fullrank", "fixed_param")
+  c("sampling", "meanfield", "fullrank", "pathfinder", "laplace", "fixed_param")
 }
 
 # check if the model was fit the the required backend
@@ -453,6 +488,10 @@ require_backend <- function(backend, x) {
 #'   \code{reduce_sum}? Defaults to \code{FALSE}. Setting it to \code{TRUE}
 #'   is required to achieve exact reproducibility of the model results
 #'   (if the random seed is set as well).
+#' @param force Logical. Defaults to \code{FALSE}. If \code{TRUE}, this will
+#'   force the Stan model to compile with threading enabled without altering the
+#'   Stan code generated by brms. This can be useful if your own custom Stan
+#'   functions use threading internally.
 #'
 #' @return A \code{brmsthreads} object which can be passed to the
 #'   \code{threads} argument of \code{brm} and related functions.
@@ -481,7 +520,8 @@ require_backend <- function(backend, x) {
 #' }
 #'
 #' @export
-threading <- function(threads = NULL, grainsize = NULL, static = FALSE) {
+threading <- function(threads = NULL, grainsize = NULL, static = FALSE,
+                      force = FALSE) {
   out <- list(threads = NULL, grainsize = NULL)
   class(out) <- "brmsthreads"
   if (!is.null(threads)) {
@@ -499,6 +539,7 @@ threading <- function(threads = NULL, grainsize = NULL, static = FALSE) {
     out$grainsize <- grainsize
   }
   out$static <- as_one_logical(static)
+  out$force <- as_one_logical(force)
   out
 }
 
@@ -521,8 +562,14 @@ validate_threads <- function(threads) {
 }
 
 # is threading activated?
-use_threading <- function(threads) {
-  isTRUE(validate_threads(threads)$threads > 0)
+use_threading <- function(threads, force = FALSE) {
+  threads <- validate_threads(threads)
+  out <- isTRUE(threads$threads > 0)
+  if (!force) {
+    # Stan code will only be altered in non-forced mode
+    out <- out && !isTRUE(threads$force)
+  }
+  out
 }
 
 #' GPU support in Stan via OpenCL
@@ -632,25 +679,72 @@ file_refit_options <- function() {
 }
 
 # canonicalize Stan model file in accordance with the current Stan version
-.canonicalize_stan_model <- function(stan_file, overwrite_file = TRUE) {
-  cmdstan_mod <- cmdstanr::cmdstan_model(stan_file, compile = FALSE)
-  out <- utils::capture.output(
-    cmdstan_mod$format(
-      canonicalize = list("deprecations", "braces", "parentheses"),
-      overwrite_file = overwrite_file, backup = FALSE
-    )
-  )
-  paste0(out, collapse = "\n")
-}
+# this function may no longer be needed due to rstan 2.26+ now being on CRAN
+# for more details see https://github.com/paul-buerkner/brms/issues/1544
+# .canonicalize_stan_model <- function(stan_file, overwrite_file = TRUE) {
+#   cmdstan_mod <- cmdstanr::cmdstan_model(stan_file, compile = FALSE)
+#   out <- utils::capture.output(
+#     cmdstan_mod$format(
+#       canonicalize = list("deprecations", "braces", "parentheses"),
+#       overwrite_file = overwrite_file, backup = FALSE
+#     )
+#   )
+#   paste0(out, collapse = "\n")
+# }
 
-# read in stan CSVs via cmdstanr and repackage into a stanfit object
-# efficient replacement of rstan::read_stan_csv
-# @param files character vector of CSV files names where draws are stored
-# @param variables character vectors of variables to extract draws for
-# @param sampler_diagnostics character vectors of diagnostics to extract
-# @return a stanfit object
-read_csv_as_stanfit <- function(files, variables = NULL,
-                                sampler_diagnostics = NULL) {
+#' Read CmdStan CSV files as a brms-formatted stanfit object
+#'
+#' \code{read_csv_as_stanfit} is used internally to read CmdStan CSV files into a
+#' \code{stanfit} object that is consistent with the structure of the fit slot of a
+#' brmsfit object.
+#'
+#' @param files Character vector of CSV files names where draws are stored.
+#' @param variables Character vector of variables to extract from the CSV files.
+#' @param sampler_diagnostics Character vector of sampler diagnostics to extract.
+#' @param model A compiled cmdstanr model object (optional). Provide this argument
+#'  if you want to allow updating the model without recompilation.
+#' @param exclude Character vector of variables to exclude from the stanfit. Only
+#'  used when \code{variables} is also specified.
+#' @param algorithm The algorithm with which the model was fitted.
+#'   See \code{\link{brm}} for details.
+#'
+#' @return A stanfit object consistent with the structure of the \code{fit}
+#'  slot of a brmsfit object.
+#'
+#' @examples
+#' \dontrun{
+#' # fit a model manually via cmdstanr
+#' scode <- stancode(count ~ Trt, data = epilepsy)
+#' sdata <- standata(count ~ Trt, data = epilepsy)
+#' mod <- cmdstanr::cmdstan_model(cmdstanr::write_stan_file(scode))
+#' stanfit <- mod$sample(data = sdata)
+#'
+#' # feed the Stan model back into brms
+#' fit <- brm(count ~ Trt, data = epilepsy, empty = TRUE, backend = 'cmdstanr')
+#' fit$fit <- read_csv_as_stanfit(stanfit$output_files(), model = mod)
+#' fit <- rename_pars(fit)
+#' summary(fit)
+#' }
+#'
+#' @export
+read_csv_as_stanfit <- function(files, variables = NULL, sampler_diagnostics = NULL,
+                                model = NULL, exclude = "", algorithm = "sampling") {
+  require_package("cmdstanr")
+
+  if (!is.null(variables)) {
+    # ensure that only relevant variables are read from CSV
+    variables <- repair_variable_names(variables)
+    variables <- unique(sub("\\[.+", "", variables))
+    variables <- setdiff(variables, exclude)
+    # temp fix for cmdstanr not recognizing the variable names it produces #1473
+    if (algorithm %in% c("meanfield", "fullrank")) {
+      variables <- ifelse(variables == "lp_approx__", "log_g__", variables)
+    } else if (algorithm %in% "pathfinder") {
+      variables <- setdiff(variables, "lp_approx__")
+    } else if (algorithm %in% "laplace") {
+      variables <- setdiff(variables, c("lp__", "lp_approx__"))
+    }
+  }
 
   csfit <- cmdstanr::read_cmdstan_csv(
     files = files, variables = variables,
@@ -662,28 +756,23 @@ read_csv_as_stanfit <- function(files, variables = NULL,
   model_name = gsub(".csv", "", basename(files[[1]]))
 
   # @model_pars
-  svars <- csfit$metadata$stan_variables
+  model_pars <- csfit$metadata$stan_variables
   if (!is.null(variables)) {
-    variables_main <- unique(gsub("\\[.*\\]", "", variables))
-    svars <- intersect(variables_main, svars)
+    model_pars <- intersect(model_pars, variables)
   }
-  if ("lp__" %in% svars) {
-    svars <- c(setdiff(svars, "lp__"), "lp__")
-  }
-  pars_oi <- svars
-  par_names <- csfit$metadata$model_params
+  # special variables will be added back later on
+  special_vars <- c("lp__", "lp_approx__")
+  model_pars <- setdiff(model_pars, special_vars)
 
   # @par_dims
-  par_dims <- vector("list", length(svars))
-
-  names(par_dims) <- svars
-  par_dims <- lapply(par_dims, function(x) x <- integer(0))
-
-  pdims_num <- ulapply(
-    svars, function(x) sum(grepl(paste0("^", x, "\\[.*\\]$"), par_names))
+  par_dims <- vector("list", length(model_pars))
+  names(par_dims) <- model_pars
+  par_dims <- lapply(par_dims, function(x) integer(0))
+  pdims_num <- ulapply(model_pars, function(x)
+    sum(grepl(paste0("^", x, "\\[.*\\]$"), csfit$metadata$model_params))
   )
   par_dims[pdims_num != 0] <-
-    csfit$metadata$stan_variable_sizes[svars][pdims_num != 0]
+    csfit$metadata$stan_variable_sizes[model_pars][pdims_num != 0]
 
   # @mode
   mode <- 0L
@@ -723,8 +812,11 @@ read_csv_as_stanfit <- function(files, variables = NULL,
     csfit$post_warmup_draws <- NULL
 
     # prepare sampler diagnostics
-    diagnostics <- rbind(csfit$warmup_sampler_diagnostics,
-                         csfit$post_warmup_sampler_diagnostics)
+    diagnostics <- rbind(
+      csfit$warmup_sampler_diagnostics,
+      csfit$post_warmup_sampler_diagnostics
+    )
+
     # manage memory
     csfit$warmup_sampler_diagnostics <- NULL
     csfit$post_warmup_sampler_diagnostics <- NULL
@@ -752,20 +844,25 @@ read_csv_as_stanfit <- function(files, variables = NULL,
     diagnostics <- as.data.frame(matrix(nrow = nrow(samples), ncol = 0))
   }
 
+  # some diagnostics may be missing in the output depending on the algorithm
+  rstan_diagn_order <- intersect(rstan_diagn_order, names(diagnostics))
+
   # convert to regular data.frame
   samples <- as.data.frame(samples)
   chain_ids <- samples$.chain
   samples[res_vars] <- NULL
-  if ("lp__" %in% colnames(samples)) {
-    samples <- move2end(samples, "lp__")
+
+  # only add special variables to dims if there are present in samples
+  # this ensures that dims_oi, pars_oi, and fnames_oi match with samples
+  for (p in special_vars) {
+    if (p %in% colnames(samples)) {
+      samples <- move2end(samples, p)
+      par_dims[[p]] <- integer(0)
+    }
   }
-
+  model_pars <- names(par_dims)
   fnames_oi <- colnames(samples)
 
-  colnames(samples) <- gsub("\\[", ".", colnames(samples))
-  colnames(samples) <- gsub("\\]", "", colnames(samples))
-  colnames(samples) <- gsub("\\,", ".", colnames(samples))
-
   # split samples into chains
   samples <- split(samples, chain_ids)
   names(samples) <- NULL
@@ -803,27 +900,31 @@ read_csv_as_stanfit <- function(files, variables = NULL,
   idx_samples <- (n_iter_warmup + 1):(n_iter_warmup + n_iter_sample)
 
   for (i in seq_along(samples)) {
-    m <- colMeans(samples[[i]][idx_samples, , drop=FALSE])
+    m <- colMeans(samples[[i]][idx_samples, , drop = FALSE])
     rownames(samples[[i]]) <- seq_rows(samples[[i]])
     attr(samples[[i]], "sampler_params") <- diagnostics[[i]][rstan_diagn_order]
     rownames(attr(samples[[i]], "sampler_params")) <- seq_rows(diagnostics[[i]])
 
     # reformat back to text
-    if (is_equal(sampler_t, "NUTS(dense_e)")) {
-      mmatrix_txt <- "\n# Elements of inverse mass matrix:\n# "
-      mmat <- paste0(apply(csfit$inv_metric[[i]], 1, paste0, collapse=", "),
-                     collapse="\n# ")
+    if (length(csfit$inv_metric)) {
+      if (is_equal(sampler_t, "NUTS(dense_e)")) {
+        mmatrix_txt <- "\n# Elements of inverse mass matrix:\n# "
+        mmat <- paste0(apply(csfit$inv_metric[[i]], 1, paste0, collapse = ", "),
+                       collapse = "\n# ")
+      } else {
+        mmatrix_txt <- "\n# Diagonal elements of inverse mass matrix:\n# "
+        mmat <- paste0(csfit$inv_metric[[i]], collapse = ", ")
+      }
+
+      adapt_info[[i]] <- paste0("# Step size = ",
+                                csfit$step_size[[i]],
+                                mmatrix_txt,
+                                mmat, "\n# ")
+      attr(samples[[i]], "adaptation_info") <- adapt_info[[i]]
     } else {
-      mmatrix_txt <- "\n# Diagonal elements of inverse mass matrix:\n# "
-      mmat <- paste0(csfit$inv_metric[[i]], collapse = ", ")
+      attr(samples[[i]], "adaptation_info") <- character(0)
     }
 
-    adapt_info[[i]] <- paste0("# Step size = ",
-                              csfit$step_size[[i]],
-                              mmatrix_txt,
-                              mmat, "\n# ")
-
-    attr(samples[[i]], "adaptation_info") <- adapt_info[[i]]
 
     attr(samples[[i]], "args") <- list(sampler_t = sampler_t, chain_id = i)
 
@@ -849,7 +950,7 @@ read_csv_as_stanfit <- function(files, variables = NULL,
     n_save = rep(n_iter_sample + n_iter_warmup, n_chains),
     warmup2 = rep(n_iter_warmup, n_chains),
     permutation = perm_lst,
-    pars_oi = pars_oi,
+    pars_oi = model_pars,
     dims_oi = par_dims,
     fnames_oi = fnames_oi,
     n_flatnames = length(fnames_oi)
@@ -923,8 +1024,10 @@ read_csv_as_stanfit <- function(files, variables = NULL,
   }
 
   # @stanmodel
+  cxxdso_class <- "cxxdso"
+  attr(cxxdso_class, "package") <- "rstan"
   null_dso <- new(
-    "cxxdso", sig = list(character(0)), dso_saved = FALSE,
+    cxxdso_class, sig = list(character(0)), dso_saved = FALSE,
     dso_filename = character(0), modulename = character(0),
     system = R.version$system, cxxflags = character(0),
     .CXXDSOMISC = new.env(parent = emptyenv())
@@ -938,10 +1041,10 @@ read_csv_as_stanfit <- function(files, variables = NULL,
   sdate <- do.call(max, lapply(files, function(csv) file.info(csv)$mtime))
   sdate <- format(sdate, "%a %b %d %X %Y")
 
-  new(
+  out <- new(
     "stanfit",
     model_name = model_name,
-    model_pars = svars,
+    model_pars = model_pars,
     par_dims = par_dims,
     mode = mode,
     sim = sim,
@@ -951,4 +1054,8 @@ read_csv_as_stanfit <- function(files, variables = NULL,
     date = sdate,  # not the time of sampling
     .MISC = new.env(parent = emptyenv())
   )
+
+  attributes(out)$metadata <- csfit
+  attributes(out)$CmdStanModel <- model
+  out
 }
