@@ -31,7 +31,7 @@
 #'   \code{family} might also be a list of families.
 #' @param prior One or more \code{brmsprior} objects created by
 #'   \code{\link{set_prior}} or related functions and combined using the
-#'   \code{c} method or the \code{+} operator. See also  \code{\link{get_prior}}
+#'   \code{c} method or the \code{+} operator. See also \code{\link[brms:default_prior.default]{default_prior}}
 #'   for more help.
 #' @param data2 A named \code{list} of objects containing data, which
 #'   cannot be passed via argument \code{data}. Required for some objects
@@ -76,7 +76,8 @@
 #'   variables defined in Stan's \code{parameters} block should be saved
 #'   (default is \code{FALSE}). Saving these draws is required in order to
 #'   apply the methods \code{bridge_sampler}, \code{bayes_factor}, and
-#'   \code{post_prob}.
+#'   \code{post_prob}. Can be set globally for the current \R session via the
+#'   \code{"brms.save_pars"} option (see \code{\link{options}}).
 #' @param sample_prior Indicate if draws from priors should be drawn
 #'   additionally to the posterior draws. Options are \code{"no"} (the
 #'   default), \code{"yes"}, and \code{"only"}. Among others, these draws can
@@ -162,7 +163,9 @@
 #'   Options are \code{"sampling"} for MCMC (the default), \code{"meanfield"} for
 #'   variational inference with independent normal distributions,
 #'   \code{"fullrank"} for variational inference with a multivariate normal
-#'   distribution, or \code{"fixed_param"} for sampling from fixed parameter
+#'   distribution, \code{"pathfinder"} for the pathfinder algorithm,
+#'   \code{"laplace"} for the laplace approximation,
+#'   or \code{"fixed_param"} for sampling from fixed parameter
 #'   values. Can be set globally for the current \R session via the
 #'   \code{"brms.algorithm"} option (see \code{\link{options}}).
 #' @param backend Character string naming the package to use as the backend for
@@ -208,7 +211,11 @@
 #'   files won't be overwritten, you have to manually remove the file in order
 #'   to refit and save the model under an existing file name. The file name
 #'   is stored in the \code{brmsfit} object for later usage.
-#' @param file_refit Modifies when the fit stored via the \code{file} parameter
+#' @param file_compress Logical or a character string, specifying one of the
+#'   compression algorithms supported by \code{\link{saveRDS}}. If the
+#'   \code{file} argument is provided, this compression will be used when saving
+#'   the fitted model object.
+#' @param file_refit Modifies when the fit stored via the \code{file} argument
 #'   is re-used. Can be set globally for the current \R session via the
 #'   \code{"brms.file_refit"} option (see \code{\link{options}}).
 #'   For \code{"never"} (default) the fit is always loaded if it
@@ -266,7 +273,7 @@
 #'   \code{\link[brms:set_prior]{set_prior}} function. Its documentation
 #'   contains detailed information on how to correctly specify priors. To find
 #'   out on which parameters or parameter classes priors can be defined, use
-#'   \code{\link[brms:get_prior]{get_prior}}. Default priors are chosen to be
+#'   \code{\link[brms:default_prior.default]{default_prior}}. Default priors are chosen to be
 #'   non or very weakly informative so that their influence on the results will
 #'   be negligible and you usually don't have to worry about them. However,
 #'   after getting more familiar with Bayesian statistics, I recommend you to
@@ -313,18 +320,18 @@
 #' @examples
 #' \dontrun{
 #' # Poisson regression for the number of seizures in epileptic patients
-#' # using normal priors for population-level effects
-#' # and half-cauchy priors for standard deviations of group-level effects
-#' prior1 <- prior(normal(0,10), class = b) +
-#'   prior(cauchy(0,2), class = sd)
-#' fit1 <- brm(count ~ zAge + zBase * Trt + (1|patient),
-#'             data = epilepsy, family = poisson(), prior = prior1)
+#' fit1 <- brm(
+#'   count ~ zBase * Trt + (1|patient),
+#'   data = epilepsy, family = poisson(),
+#'   prior = prior(normal(0, 10), class = b) +
+#'     prior(cauchy(0, 2), class = sd)
+#' )
 #'
 #' # generate a summary of the results
 #' summary(fit1)
 #'
 #' # plot the MCMC chains as well as the posterior distributions
-#' plot(fit1, ask = FALSE)
+#' plot(fit1)
 #'
 #' # predict responses based on the fitted model
 #' head(predict(fit1))
@@ -408,13 +415,13 @@
 #'
 #' # use the future package for more flexible parallelization
 #' library(future)
-#' plan(multiprocess)
+#' plan(multisession, workers = 4)
 #' fit7 <- update(fit7, future = TRUE)
 #'
 #'
 #' # fit a model manually via rstan
-#' scode <- make_stancode(count ~ Trt, data = epilepsy)
-#' sdata <- make_standata(count ~ Trt, data = epilepsy)
+#' scode <- stancode(count ~ Trt, data = epilepsy)
+#' sdata <- standata(count ~ Trt, data = epilepsy)
 #' stanfit <- rstan::stan(model_code = scode, data = sdata)
 #' # feed the Stan model back into brms
 #' fit8 <- brm(count ~ Trt, data = epilepsy, empty = TRUE)
@@ -432,8 +439,8 @@ brm <- function(formula, data, family = gaussian(), prior = NULL,
                 autocor = NULL, data2 = NULL, cov_ranef = NULL,
                 sample_prior = "no", sparse = NULL, knots = NULL,
                 drop_unused_levels = TRUE, stanvars = NULL, stan_funs = NULL,
-                fit = NA, save_pars = NULL, save_ranef = NULL,
-                save_mevars = NULL, save_all_pars = NULL,
+                fit = NA, save_pars = getOption("brms.save_pars", NULL),
+                save_ranef = NULL, save_mevars = NULL, save_all_pars = NULL,
                 init = NULL, inits = NULL, chains = 4, iter = 2000,
                 warmup = floor(iter / 2), thin = 1,
                 cores = getOption("mc.cores", 1),
@@ -445,7 +452,8 @@ brm <- function(formula, data, family = gaussian(), prior = NULL,
                 backend = getOption("brms.backend", "rstan"),
                 future = getOption("future", FALSE), silent = 1,
                 seed = NA, save_model = NULL, stan_model_args = list(),
-                file = NULL, file_refit = getOption("brms.file_refit", "never"),
+                file = NULL, file_compress = TRUE,
+                file_refit = getOption("brms.file_refit", "never"),
                 empty = FALSE, rename = TRUE, ...) {
 
   # optionally load brmsfit from file
@@ -512,15 +520,15 @@ brm <- function(formula, data, family = gaussian(), prior = NULL,
       get_data2_autocor(formula),
       get_data2_cov_ranef(formula)
     )
-    data_name <- substitute_name(data)
     data <- validate_data(
       data, bterms = bterms,
       data2 = data2, knots = knots,
-      drop_unused_levels = drop_unused_levels
+      drop_unused_levels = drop_unused_levels,
+      data_name = substitute_name(data)
     )
-    attr(data, "data_name") <- data_name
+    bframe <- brmsframe(bterms, data)
     prior <- .validate_prior(
-      prior, bterms = bterms, data = data,
+      prior, bframe = bframe,
       sample_prior = sample_prior
     )
     stanvars <- validate_stanvars(stanvars, stan_funs = stan_funs)
@@ -529,28 +537,27 @@ brm <- function(formula, data, family = gaussian(), prior = NULL,
       save_mevars = save_mevars,
       save_all_pars = save_all_pars
     )
-    ranef <- tidy_ranef(bterms, data = data)
-    # generate Stan code
-    model <- .make_stancode(
-      bterms, data = data, prior = prior,
-      stanvars = stanvars, save_model = save_model,
-      backend = backend, threads = threads, opencl = opencl,
-      normalize = normalize
-    )
 
+    # generate Stan code
+    model <- .stancode(
+      bframe, prior = prior, stanvars = stanvars,
+      save_model = save_model, backend = backend, threads = threads,
+      opencl = opencl, normalize = normalize
+    )
     # initialize S3 object
     x <- brmsfit(
       formula = formula, data = data, data2 = data2, prior = prior,
       stanvars = stanvars, model = model, algorithm = algorithm,
       backend = backend, threads = threads, opencl = opencl,
-      save_pars = save_pars, ranef = ranef, family = family,
+      save_pars = save_pars, ranef = bframe$frame$re, family = family,
+      basis = frame_basis(bframe, data = data),
       stan_args = nlist(init, silent, control, stan_model_args, ...)
     )
-    exclude <- exclude_pars(x)
+    exclude <- exclude_pars(x, bframe = bframe)
     # generate Stan data before compiling the model to avoid
     # unnecessary compilations in case of invalid data
-    sdata <- .make_standata(
-      bterms, data = data, prior = prior, data2 = data2,
+    sdata <- .standata(
+      bframe, data = data, prior = prior, data2 = data2,
       stanvars = stanvars, threads = threads
     )
 
@@ -594,7 +601,7 @@ brm <- function(formula, data, family = gaussian(), prior = NULL,
     x <- rename_pars(x)
   }
   if (!is.null(file)) {
-    x <- write_brmsfit(x, file)
+    x <- write_brmsfit(x, file, compress = file_compress)
   }
   x
 }
