@@ -1,6 +1,5 @@
 #' Bayesian Regression Models using 'Stan'
 #'
-#' @docType package
 #' @name brms-package
 #' @aliases brms
 #'
@@ -22,8 +21,8 @@
 #' formula syntax to specify a wide range of complex Bayesian models
 #' (see \code{\link{brmsformula}} for details). Based on the supplied
 #' formulas, data, and additional information, it writes the Stan code
-#' on the fly via \code{\link{make_stancode}}, prepares the data via
-#' \code{\link{make_standata}}, and fits the model using
+#' on the fly via \code{\link[brms:stancode.default]{stancode}}, prepares the data via
+#' \code{\link[brms:standata.default]{standata}} and fits the model using
 #' \pkg{\link[rstan:rstan]{Stan}}.
 #'
 #' Subsequently, a large number of post-processing methods can be applied:
@@ -81,4 +80,4 @@
 #' \code{\link{brmsfamily}},
 #' \code{\link{brmsfit}}
 #'
-NULL
+"_PACKAGE"
