@@ -11,6 +11,7 @@ is_mv <- function(x) {
 }
 
 stopifnot_resp <- function(x, resp = NULL) {
+  # TODO: merge into validate_resp?
   if (is_mv(x) && length(resp) != 1L) {
     stop2("Argument 'resp' must be a single variable name ",
           "when applying this method to a multivariate model.")
@@ -214,16 +215,17 @@ get_cor_matrix <- function(cor, size = NULL, ndraws = NULL) {
 # compute covariance matrices of autocor structures
 # @param prep a brmsprep object
 # @param obs observations for which to compute the covariance matrix
+# @param Jtime vector indicating to which time points obs belong
 # @param latent compute covariance matrix for latent residuals?
-get_cov_matrix_ac <- function(prep, obs = NULL, latent = FALSE) {
+get_cov_matrix_ac <- function(prep, obs = NULL, Jtime = NULL, latent = FALSE) {
   if (is.null(obs)) {
     obs <- seq_len(prep$nobs)
   }
   nobs <- length(obs)
   ndraws <- prep$ndraws
-  acef <- prep$ac$acef
+  acframe <- prep$ac$acframe
   # prepare correlations
-  if (has_ac_class(acef, "arma")) {
+  if (has_ac_class(acframe, "arma")) {
     ar <- as.numeric(prep$ac$ar)
     ma <- as.numeric(prep$ac$ma)
     if (length(ar) && !length(ma)) {
@@ -235,10 +237,13 @@ get_cov_matrix_ac <- function(prep, obs = NULL, latent = FALSE) {
     } else {
       stop2("Neither 'ar' nor 'ma' were supplied. Please report a bug.")
     }
-  } else if (has_ac_class(acef, "cosy")) {
+  } else if (has_ac_class(acframe, "cosy")) {
     cosy <- as.numeric(prep$ac$cosy)
     cor <- get_cor_matrix_cosy(cosy, nobs)
-  } else if (has_ac_class(acef, "fcor")) {
+  } else if (has_ac_class(acframe, "unstr")) {
+    cortime <- prep$ac$cortime
+    cor <- get_cor_matrix_unstr(cortime, Jtime)
+  } else if (has_ac_class(acframe, "fcor")) {
     cor <- get_cor_matrix_fcor(prep$ac$Mfcor, ndraws)
   } else {
     cor <- get_cor_matrix_ident(ndraws, nobs)
@@ -347,6 +352,17 @@ get_cor_matrix_cosy <- function(cosy, nobs) {
   out
 }
 
+# compute unstructured time correlation matrices
+# @param cortime time correlation draws
+# @param Jtime indictor of rows/cols to consider in cortime
+# @return a numeric 'ndraws' x 'nobs' x 'nobs' array
+#   where nobs = length(Jtime[Jtime > 0])
+get_cor_matrix_unstr <- function(cortime, Jtime) {
+  stopifnot(length(Jtime) > 0L)
+  Jtime <- Jtime[Jtime > 0]
+  get_cor_matrix(cortime)[, Jtime, Jtime, drop = FALSE]
+}
+
 # prepare a fixed correlation matrix
 # @param Mfcor correlation matrix to be prepared
 # @param ndraws number of posterior draws
@@ -654,7 +670,11 @@ apply_dpar_inv_link <- function(dpar, family) {
 # in categorical-like models using the softmax response function
 insert_refcat <- function(eta, refcat = 1) {
   stopifnot(is.array(eta))
-  refcat <- as_one_integer(refcat)
+  refcat <- as_one_integer(refcat, allow_na = TRUE)
+  if (isNA(refcat)) {
+    # no reference category used
+    return(eta)
+  }
   # need to add zeros for the reference category
   ndim <- length(dim(eta))
   dim_noncat <- dim(eta)[-ndim]
@@ -713,7 +733,7 @@ split_dots <- function(x, ..., model_names = NULL, other = TRUE) {
   other <- as_one_logical(other)
   dots <- list(x, ...)
   names <- substitute(list(x, ...), env = parent.frame())[-1]
-  names <- ulapply(names, deparse_combine)
+  names <- ulapply(names, deparse0)
   if (length(names)) {
     if (!length(names(dots))) {
       names(dots) <- names
@@ -745,7 +765,7 @@ split_dots <- function(x, ..., model_names = NULL, other = TRUE) {
 # @return the 'eta' matrix with possibly reordered columns
 reorder_obs <- function(eta, old_order = NULL, sort = FALSE) {
   stopifnot(length(dim(eta)) %in% c(2L, 3L))
-  if (is.null(old_order) || sort) {
+  if (!length(old_order) || sort) {
     return(eta)
   }
   stopifnot(length(old_order) == NCOL(eta))
@@ -756,11 +776,12 @@ reorder_obs <- function(eta, old_order = NULL, sort = FALSE) {
 # allows to call log_prob and other C++ using methods
 # on objects not created in the current R session
 # or objects created via another backend
-update_misc_env <- function(x, only_windows = FALSE) {
+update_misc_env <- function(x, recompile = FALSE, only_windows = FALSE) {
   stopifnot(is.brmsfit(x))
+  recompile <- as_one_logical(recompile)
   only_windows <- as_one_logical(only_windows)
-  if (!has_rstan_model(x)) {
-    x <- add_rstan_model(x)
+  if (recompile || !has_rstan_model(x)) {
+    x <- add_rstan_model(x, overwrite = TRUE)
   } else if (os_is_windows() || !only_windows) {
     # TODO: detect when updating .MISC is not required
     # TODO: find a more efficient way to update .MISC
@@ -850,9 +871,9 @@ validate_cores_post_processing <- function(cores) {
 #' and scripts should not use it.
 #'
 #' @param fit Old \code{brmsfit} object (e.g., loaded from file).
-#' @param sdata New Stan data (result of a call to \code{\link{make_standata}}).
+#' @param sdata New Stan data (result of a call to \code{\link[brms:standata.default]{standata}}).
 #'   Pass \code{NULL} to avoid this data check.
-#' @param scode New Stan code (result of a call to \code{\link{make_stancode}}).
+#' @param scode New Stan code (result of a call to \code{\link[brms:stancode.default]{stancode}}).
 #'   Pass \code{NULL} to avoid this code check.
 #' @param data New data to check consistency of factor level names.
 #'   Pass \code{NULL} to avoid this data check.
@@ -972,7 +993,7 @@ read_brmsfit <- function(file) {
     )
   }
   x <- suppressWarnings(try(readRDS(file), silent = TRUE))
-  if (!is(x, "try-error")) {
+  if (!is_try_error(x)) {
     if (!is.brmsfit(x)) {
       stop2("Object loaded via 'file' is not of class 'brmsfit'.")
     }
@@ -986,12 +1007,13 @@ read_brmsfit <- function(file) {
 # write a brmsfit object to a file
 # @param x a brmsfit object
 # @param file path to an rds file
+# @param compress compression format supported by saveRDS
 # @return NULL
-write_brmsfit <- function(x, file) {
+write_brmsfit <- function(x, file, compress = TRUE) {
   stopifnot(is.brmsfit(x))
   file <- check_brmsfit_file(file)
   x$file <- file
-  saveRDS(x, file = file)
+  saveRDS(x, file = file, compress = compress)
   invisible(x)
 }
 
