@@ -64,13 +64,12 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
   y$cov_ranef <- x$cov_ranef
   class(y) <- "brmsterms"
 
+  y$resp <- ""
   if (check_response) {
     # extract response variables
     y$respform <- validate_resp_formula(formula, empty_ok = FALSE)
     if (mv) {
       y$resp <- terms_resp(y$respform)
-    } else {
-      y$resp <- ""
     }
   }
 
@@ -97,6 +96,11 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
       x$pforms[[dp]] <- combine_formulas(formula, x$pforms[[dp]], dp)
     }
     x$pforms <- move2start(x$pforms, mu_dpars)
+    for (i in seq_along(family$mix)) {
+      # store the respective mixture index in each mixture component
+      # this enables them to be easily passed along, e.g. in stan_log_lik
+      y$family$mix[[i]]$mix <- i
+    }
   } else if (conv_cats_dpars(x$family)) {
     mu_dpars <- str_subset(x$family$dpars, "^mu")
     for (dp in mu_dpars) {
@@ -109,7 +113,6 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
   }
 
   # predicted distributional parameters
-  resp <- ifelse(mv && !is.null(y$resp), y$resp, "")
   dpars <- intersect(names(x$pforms), valid_dpars(family))
   dpar_forms <- x$pforms[dpars]
   nlpars <- setdiff(names(x$pforms), dpars)
@@ -117,17 +120,18 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
   y$dpars <- named_list(dpars)
   for (dp in dpars) {
     if (get_nl(dpar_forms[[dp]])) {
-      y$dpars[[dp]] <- terms_nlf(dpar_forms[[dp]], nlpars, resp)
+      y$dpars[[dp]] <- terms_nlf(dpar_forms[[dp]], nlpars, y$resp)
     } else {
       y$dpars[[dp]] <- terms_lf(dpar_forms[[dp]])
     }
     y$dpars[[dp]]$family <- dpar_family(family, dp)
     y$dpars[[dp]]$dpar <- dp
-    y$dpars[[dp]]$resp <- resp
+    y$dpars[[dp]]$resp <- y$resp
     if (dpar_class(dp) == "mu") {
       y$dpars[[dp]]$respform <- y$respform
       y$dpars[[dp]]$adforms <- y$adforms
     }
+    y$dpars[[dp]]$transform <- stan_eta_transform(y, y$dpars[[dp]]$family)
     check_cs(y$dpars[[dp]])
   }
 
@@ -141,15 +145,15 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
         attr(nlpar_forms[[nlp]], "center") <- FALSE
       }
       if (get_nl(nlpar_forms[[nlp]])) {
-        y$nlpars[[nlp]] <- terms_nlf(nlpar_forms[[nlp]], nlpars, resp)
+        y$nlpars[[nlp]] <- terms_nlf(nlpar_forms[[nlp]], nlpars, y$resp)
       } else {
         y$nlpars[[nlp]] <- terms_lf(nlpar_forms[[nlp]])
       }
       y$nlpars[[nlp]]$nlpar <- nlp
-      y$nlpars[[nlp]]$resp <- resp
+      y$nlpars[[nlp]]$resp <- y$resp
       check_cs(y$nlpars[[nlp]])
     }
-    used_nlpars <- ulapply(c(y$dpars, y$nlpars), "[[", "used_nlpars")
+    used_nlpars <- ufrom_list(c(y$dpars, y$nlpars), "used_nlpars")
     unused_nlpars <- setdiff(nlpars, used_nlpars)
     if (length(unused_nlpars)) {
       stop2(
@@ -159,7 +163,7 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
       )
     }
     # sort non-linear parameters after dependency
-    used_nlpars <- lapply(y$nlpars, "[[", "used_nlpars")
+    used_nlpars <- from_list(y$nlpars, "used_nlpars")
     sorted_nlpars <- sort_dependencies(used_nlpars)
     y$nlpars <- y$nlpars[sorted_nlpars]
   }
@@ -201,16 +205,17 @@ brmsterms.brmsformula <- function(formula, check_response = TRUE,
   y$allvars <- allvars_formula(
     lhsvars, advars, lapply(y$dpars, get_allvars),
     lapply(y$nlpars, get_allvars), y$time$allvars,
-    get_unused_arg_vars(y)
+    get_unused_arg_vars(y),
+    .env = environment(formula)
   )
   if (check_response) {
     # add y$respform to the left-hand side of y$allvars
     # avoid using update.formula as it is inefficient for longer formulas
     formula_allvars <- y$respform
     formula_allvars[[3]] <- y$allvars[[2]]
+    environment(formula_allvars) <- environment(y$allvars)
     y$allvars <- formula_allvars
   }
-  environment(y$allvars) <- environment(formula)
   y
 }
 
@@ -228,12 +233,15 @@ brmsterms.mvbrmsformula <- function(formula, ...) {
     x$forms[[i]]$mv <- TRUE
     out$terms[[i]] <- brmsterms(x$forms[[i]], ...)
   }
-  out$allvars <- allvars_formula(lapply(out$terms, get_allvars))
+  list_allvars <- lapply(out$terms, get_allvars)
+  out$allvars <- allvars_formula(
+    list_allvars, .env = environment(list_allvars[[1]])
+  )
   # required to find variables used solely in the response part
-  lhs_resp <- function(x) deparse_combine(lhs(x$respform)[[2]])
+  lhs_resp <- function(x) deparse0(lhs(x$respform)[[2]])
   out$respform <- paste0(ulapply(out$terms, lhs_resp), collapse = ",")
   out$respform <- formula(paste0("mvbind(", out$respform, ") ~ 1"))
-  out$responses <- ulapply(out$terms, "[[", "resp")
+  out$responses <- ufrom_list(out$terms, "resp")
   out$rescor <- x$rescor
   out$mecor <- x$mecor
   out$cov_ranef <- x$cov_ranef
@@ -261,7 +269,6 @@ terms_lf <- function(formula) {
     get_allvars(y$sm), get_allvars(y$gp),
     get_allvars(y$ac), get_allvars(y$offset)
   )
-  environment(y$allvars) <- environment(formula)
   structure(y, class = "btl")
 }
 
@@ -285,7 +292,6 @@ terms_nlf <- function(formula, nlpars, resp = "") {
     y$ac <- terms_ac(attr(formula, "autocor"))
   }
   y$allvars <- allvars_formula(covars, get_allvars(y$ac))
-  environment(y$allvars) <- environment(formula)
   y$loop <- loop
   structure(y, class = "btnl")
 }
@@ -433,9 +439,9 @@ terms_sp <- function(formula) {
   if (!length(out)) {
     return(NULL)
   }
-  uni_mo <- trim_wsp(get_matches_expr(regex_sp("mo"), out))
-  uni_me <- trim_wsp(get_matches_expr(regex_sp("me"), out))
-  uni_mi <- trim_wsp(get_matches_expr(regex_sp("mi"), out))
+  uni_mo <- get_matches_expr(regex_sp("mo"), out)
+  uni_me <- get_matches_expr(regex_sp("me"), out)
+  uni_mi <- get_matches_expr(regex_sp("mi"), out)
   # remove the intercept as it is handled separately
   out <- str2formula(c("0", out))
   attr(out, "int") <- FALSE
@@ -470,8 +476,8 @@ terms_gp <- function(formula) {
     return(NULL)
   }
   eterms <- lapply(out, eval2, envir = environment())
-  covars <- lapply(eterms, "[[", "term")
-  byvars <- lapply(eterms, "[[", "by")
+  covars <- from_list(eterms, "term")
+  byvars <- from_list(eterms, "by")
   allvars <- str2formula(unlist(c(covars, byvars)))
   allvars <- str2formula(all_vars(allvars))
   if (!length(all_vars(allvars))) {
@@ -491,8 +497,8 @@ terms_ac <- function(formula) {
   }
   eterms <- lapply(out, eval2, envir = environment())
   allvars <- unlist(c(
-    lapply(eterms, "[[", "time"),
-    lapply(eterms, "[[", "gr")
+    from_list(eterms, "time"),
+    from_list(eterms, "gr")
   ))
   allvars <- str2formula(all_vars(allvars))
   out <- str2formula(out)
@@ -510,7 +516,7 @@ terms_offset <- function(formula) {
     return(NULL)
   }
   vars <- attr(formula, "variables")
-  out <- ulapply(pos, function(i) deparse(vars[[i + 1]]))
+  out <- ulapply(pos, function(i) deparse0(vars[[i + 1]]))
   out <- str2formula(out)
   attr(out, "allvars") <- str2formula(all_vars(out))
   out
@@ -576,6 +582,11 @@ is.mvbrmsterms <- function(x) {
   inherits(x, "mvbrmsterms")
 }
 
+# useful for functions that require either of the two objects
+is.anybrmsterms <- function(x) {
+  is.brmsterms(x) || is.mvbrmsterms(x)
+}
+
 is.btl <- function(x) {
   inherits(x, "btl")
 }
@@ -584,6 +595,16 @@ is.btnl <- function(x) {
   inherits(x, "btnl")
 }
 
+# figure out if a certain distributional parameter is predicted
+is_pred_dpar <- function(bterms, dpar) {
+  stopifnot(is.brmsterms(bterms))
+  if (!length(dpar)) {
+    return(FALSE)
+  }
+  mix <- get_mix_id(bterms)
+  any(paste0(dpar, mix) %in% names(bterms$dpars))
+}
+
 # transform mvbrmsterms objects for use in stan_llh.brmsterms
 as.brmsterms <- function(x) {
   stopifnot(is.mvbrmsterms(x), x$rescor)
@@ -591,13 +612,14 @@ as.brmsterms <- function(x) {
   stopifnot(all(families == families[1]))
   out <- structure(list(), class = "brmsterms")
   out$family <- structure(
-    list(family = paste0(families[1], "_mv"), link = "identity"),
+    list(family = families[1], link = "identity"),
     class = c("brmsfamily", "family")
   )
+  out$family$fun <- paste0(out$family$family, "_mv")
   info <- get(paste0(".family_", families[1]))()
   out$family[names(info)] <- info
   out$sigma_pred <- any(ulapply(x$terms,
-    function(x) "sigma" %in% names(x$dpar) || is.formula(x$adforms$se)
+    function(x) is_pred_dpar(x, "sigma") || has_ad_terms(x, "se")
   ))
   weight_forms <- rmNULL(lapply(x$terms, function(x) x$adforms$weights))
   if (length(weight_forms)) {
@@ -704,7 +726,7 @@ check_fdpars <- function(x) {
 # combine all variables in one formuula
 # @param x (list of) formulas or character strings
 # @return a formula with all variables on the right-hand side
-allvars_formula <- function(...) {
+allvars_formula <- function(..., .env = parent.frame()) {
   out <- rmNULL(c(...))
   out <- collapse(ulapply(out, plus_rhs))
   all_vars <- all_vars(out)
@@ -713,7 +735,7 @@ allvars_formula <- function(...) {
     stop2("The following variable names are invalid: ",
           collapse_comma(invalid_vars))
   }
-  str2formula(c(out, all_vars))
+  str2formula(c(out, all_vars), env = .env)
 }
 
 # conveniently extract a formula of all relevant variables
@@ -802,7 +824,7 @@ combine_formulas <- function(formula1, formula2, lhs = "", update = FALSE) {
 has_terms <- function(formula) {
   stopifnot(is.formula(formula))
   terms <- try(terms(rhs(formula)), silent = TRUE)
-  is(terms, "try-error") ||
+  is_try_error(terms) ||
     length(attr(terms, "term.labels")) ||
     length(attr(terms, "offset"))
 }
@@ -903,7 +925,7 @@ get_effect.btl <- function(x, target = "fe", ...) {
 
 #' @export
 get_effect.btnl <- function(x, target = "fe", ...) {
-  NULL
+  x[[target]]
 }
 
 all_terms <- function(x) {
@@ -918,13 +940,14 @@ all_terms <- function(x) {
 
 # generate a regular expression to extract special terms
 # @param type one or more special term types to be extracted
+# TODO: rule out expressions such as mi(y) + mi(x)
 regex_sp <- function(type = "all") {
   choices <- c("all", "sp", "sm", "gp", "cs", "mmc", "ac", all_sp_types())
   type <- unique(match.arg(type, choices, several.ok = TRUE))
   funs <- c(
     sm = "(s|(t2)|(te)|(ti))",
     gp = "gp", cs = "cse?", mmc = "mmc",
-    ac = "((arma)|(ar)|(ma)|(cosy)|(sar)|(car)|(fcor))"
+    ac = "((arma)|(ar)|(ma)|(cosy)|(unstr)|(sar)|(car)|(fcor))"
   )
   funs[all_sp_types()] <- all_sp_types()
   if ("sp" %in% type) {
@@ -953,7 +976,7 @@ find_terms <- function(x, type, complete = TRUE, ranef = FALSE) {
   if (is.formula(x)) {
     x <- all_terms(x)
   } else {
-    x <- as.character(x)
+    x <- trim_wsp(as.character(x))
   }
   complete <- as_one_logical(complete)
   ranef <- as_one_logical(ranef)
@@ -966,13 +989,14 @@ find_terms <- function(x, type, complete = TRUE, ranef = FALSE) {
   if (complete) {
     matches <- lapply(out, get_matches_expr, pattern = regex)
     # each term may contain only one special function call
-    inv <- out[lengths(matches) > 1L]
-    if (!length(inv)) {
+    invalid <- out[lengths(matches) > 1L]
+    if (!length(invalid)) {
       # each term must be exactly equal to the special function call
-      inv <- out[trim_wsp(unlist(matches)) != out]
+      invalid <- out[unlist(matches) != out]
     }
-    if (length(inv)) {
-      stop2("The term '", inv[1], "' is invalid in brms syntax.")
+    # TODO: some terms can be part of I() calls (#1520); reflect this here?
+    if (length(invalid)) {
+      stop2("The term '", invalid[1], "' is invalid in brms syntax.")
     }
   }
   out
@@ -1007,7 +1031,7 @@ has_intercept <- function(formula) {
   } else {
     formula <- as.formula(formula)
     try_terms <- try(terms(formula), silent = TRUE)
-    if (is(try_terms, "try-error")) {
+    if (is_try_error(try_terms)) {
       out <- FALSE
     } else {
       out <- as.logical(attr(try_terms, "intercept"))
@@ -1033,12 +1057,12 @@ has_rsv_intercept <- function(formula, has_intercept = NULL) {
     return(.has_rsv_intercept(formula, has_intercept))
   }
   formula <- try(as.formula(formula), silent = TRUE)
-  if (is(formula, "try-error")) {
+  if (is_try_error(formula)) {
     return(FALSE)
   }
   if (is.null(has_intercept)) {
     try_terms <- try(terms(formula), silent = TRUE)
-    if (is(try_terms, "try-error")) {
+    if (is_try_error(try_terms)) {
       return(FALSE)
     }
     has_intercept <- has_intercept(try_terms)
