@@ -9,10 +9,11 @@
 #   and (2) that the number of variables matches the number
 #   of variable names; fixes issue #73
 # @param knots: a list of knot values for GAMMs
+# @param data_name: optional name of the data frame as passed by the user
 # @return model.frame for use in brms functions
 validate_data <- function(data, bterms, data2 = list(), knots = NULL,
                           na_action = na_omit, drop_unused_levels = TRUE,
-                          attr_terms = NULL) {
+                          attr_terms = NULL, data_name = "") {
   if (missing(data)) {
     stop2("Data must be specified using the 'data' argument.")
   }
@@ -20,7 +21,7 @@ validate_data <- function(data, bterms, data2 = list(), knots = NULL,
     knots <- get_knots(data)
   }
   data <- try(as.data.frame(data), silent = TRUE)
-  if (is(data, "try-error")) {
+  if (is_try_error(data)) {
     stop2("Argument 'data' must be coercible to a data.frame.")
   }
   if (!isTRUE(nrow(data) > 0L)) {
@@ -50,11 +51,14 @@ validate_data <- function(data, bterms, data2 = list(), knots = NULL,
   attributes(all_vars_terms)[names(attr_terms)] <- attr_terms
   # 'terms' prevents correct validation in 'model.frame'
   attr(data, "terms") <- NULL
+  # ensures that na_action can be passed to model.frame
+  na_action_bterms <- function(object, ...) {
+    na_action(object, bterms = bterms, ...)
+  }
   data <- model.frame(
-    all_vars_terms, data, na.action = na.pass,
+    all_vars_terms, data, na.action = na_action_bterms,
     drop.unused.levels = drop_unused_levels
   )
-  data <- na_action(data, bterms = bterms)
   if (any(grepl("__|_$", colnames(data)))) {
     stop2("Variable names may not contain double underscores ",
           "or underscores at the end.")
@@ -68,6 +72,7 @@ validate_data <- function(data, bterms, data2 = list(), knots = NULL,
   data <- fix_factor_contrasts(data, ignore = groups)
   attr(data, "knots") <- knots
   attr(data, "drop_unused_levels") <- drop_unused_levels
+  attr(data, "data_name") <- data_name
   data
 }
 
@@ -97,25 +102,25 @@ validate_data2 <- function(data2, bterms, ...) {
     }
   }
   # validate autocorrelation matrices
-  acef <- tidy_acef(bterms)
-  sar_M_names <- get_ac_vars(acef, "M", class = "sar")
+  acframe <- frame_ac(bterms)
+  sar_M_names <- get_ac_vars(acframe, "M", class = "sar")
   for (M in sar_M_names) {
     data2[[M]] <- validate_sar_matrix(get_from_data2(M, data2))
     attr(data2[[M]], "obs_based_matrix") <- TRUE
   }
-  car_M_names <- get_ac_vars(acef, "M", class = "car")
+  car_M_names <- get_ac_vars(acframe, "M", class = "car")
   for (M in car_M_names) {
     data2[[M]] <- validate_car_matrix(get_from_data2(M, data2))
     # observation based CAR matrices are deprecated and
     # there is no need to label them as observation based
   }
-  fcor_M_names <- get_ac_vars(acef, "M", class = "fcor")
+  fcor_M_names <- get_ac_vars(acframe, "M", class = "fcor")
   for (M in fcor_M_names) {
     data2[[M]] <- validate_fcor_matrix(get_from_data2(M, data2))
     attr(data2[[M]], "obs_based_matrix") <- TRUE
   }
   # validate within-group covariance matrices
-  cov_names <- ulapply(get_re(bterms)$gcall, "[[", "cov")
+  cov_names <- ufrom_list(get_re(bterms)$gcall, "cov")
   cov_names <- cov_names[nzchar(cov_names)]
   for (cov in cov_names) {
     data2[[cov]] <- validate_recov_matrix(get_from_data2(cov, data2))
@@ -264,6 +269,7 @@ subset_data <- function(data, bterms) {
     # cross-formula indexing is no longer trivial for subsetted models
     check_cross_formula_indexing(bterms)
     data <- data[subset, , drop = FALSE]
+    attr(data, "subset") <- subset
   }
   if (!NROW(data)) {
     stop2(
@@ -395,7 +401,7 @@ validate_newdata <- function(
   incl_autocor = TRUE, group_vars = NULL, req_vars = NULL, ...
 ) {
   newdata <- try(as.data.frame(newdata), silent = TRUE)
-  if (is(newdata, "try-error")) {
+  if (is_try_error(newdata)) {
     stop2("Argument 'newdata' must be coercible to a data.frame.")
   }
   object <- restructure(object)
@@ -416,7 +422,7 @@ validate_newdata <- function(
     # variables not used in the included model parts
     # do not need to be specified in newdata
     resp <- validate_resp(resp, bterms$responses)
-    form_req_vars <- lapply(bterms$terms[resp], "[[", "allvars")
+    form_req_vars <- from_list(bterms$terms[resp], "allvars")
     form_req_vars <- allvars_formula(form_req_vars)
     req_vars <- intersect(req_vars, all.vars(form_req_vars))
   }
@@ -547,16 +553,15 @@ validate_newdata <- function(
   unused_vars <- setdiff(all_vars, used_vars)
   newdata <- fill_newdata(newdata, unused_vars)
   # validate grouping factors
-  new_ranef <- tidy_ranef(bterms, data = mf)
-  new_meef <- tidy_meef(bterms, data = mf)
-  old_levels <- get_levels(new_ranef, new_meef)
+  old_levels <- get_levels(bterms, data = mf)
   if (!allow_new_levels) {
-    new_levels <- get_levels(
-      tidy_ranef(bterms, data = newdata),
-      tidy_meef(bterms, data = newdata)
-    )
+    new_levels <- get_levels(bterms, data = newdata)
     for (g in names(old_levels)) {
       unknown_levels <- setdiff(new_levels[[g]], old_levels[[g]])
+      # NA is not found by get_levels but still behaves like a new level (#1652)
+      if (anyNA(newdata[[g]])) {
+        c(unknown_levels) <- NA
+      }
       if (length(unknown_levels)) {
         unknown_levels <- collapse_comma(unknown_levels)
         stop2(
@@ -617,10 +622,13 @@ validate_newdata2 <- function(newdata2, object, ...) {
 }
 
 # extract the current data
-current_data <- function(object, newdata = NULL, ...) {
+current_data <- function(object, newdata = NULL, skip_validation = FALSE, ...) {
   stopifnot(is.brmsfit(object))
+  skip_validation <- as_one_logical(skip_validation)
   if (is.null(newdata)) {
     data <- object$data
+  } else if (skip_validation) {
+    data <- newdata
   } else {
     data <- validate_newdata(newdata, object = object, ...)
   }
@@ -628,10 +636,13 @@ current_data <- function(object, newdata = NULL, ...) {
 }
 
 # extract the current data2
-current_data2 <- function(object, newdata2 = NULL, ...) {
+current_data2 <- function(object, newdata2 = NULL, skip_validation = FALSE, ...) {
   stopifnot(is.brmsfit(object))
+  skip_validation <- as_one_logical(skip_validation)
   if (is.null(newdata2)) {
     data2 <- object$data2
+  } else if (skip_validation) {
+    data2 <- newdata2
   } else {
     data2 <- validate_newdata2(newdata2, object = object, ...)
   }
