@@ -62,11 +62,10 @@ data_response <- function(x, ...) {
 }
 
 #' @export
-data_response.mvbrmsterms <- function(x, basis = NULL, ...) {
+data_response.mvbrmsframe <- function(x, ...) {
   out <- list()
   for (i in seq_along(x$terms)) {
-    bs <- basis$resps[[x$responses[i]]]
-    c(out) <- data_response(x$terms[[i]], basis = bs, ...)
+    c(out) <- data_response(x$terms[[i]], ...)
   }
   if (x$rescor) {
     out$nresp <- length(x$responses)
@@ -76,29 +75,36 @@ data_response.mvbrmsterms <- function(x, basis = NULL, ...) {
 }
 
 #' @export
-data_response.brmsterms <- function(x, data, check_response = TRUE,
-                                    internal = FALSE, basis = NULL, ...) {
+data_response.brmsframe <- function(x, data, check_response = TRUE,
+                                    internal = FALSE, ...) {
   data <- subset_data(data, x)
   N <- nrow(data)
   # TODO: rename 'Y' to 'y'
   Y <- model.response(model.frame(x$respform, data, na.action = na.pass))
   out <- list(N = N, Y = unname(Y))
   if (is_binary(x$family)) {
-    bin_levels <- basis$resp_levels
+    bin_levels <- x$basis$resp_levels
     if (is.null(bin_levels)) {
-      bin_levels <- sort(unique(out$Y))
+      bin_levels <- levels(as.factor(out$Y))
     }
-    # fixes issue #1298
-    if (is.numeric(out$Y) && length(bin_levels) == 1L && !0 %in% bin_levels) {
-      bin_levels <- c(0, bin_levels)
+    # fixes issues #1298 and #1511
+    if (is.numeric(out$Y) && length(bin_levels) == 1L) {
+      if (0 %in% bin_levels) {
+        # 1 as default event level
+        bin_levels <- c(0, 1)
+      } else {
+        # 0 as default non-event level
+        bin_levels <- c(0, bin_levels)
+      }
     }
-    out$Y <- as.numeric(as_factor(out$Y, levels = bin_levels)) - 1
+    out$Y <- as.integer(as_factor(out$Y, levels = bin_levels)) - 1
   }
   if (is_categorical(x$family)) {
-    out$Y <- as.numeric(as_factor(out$Y, levels = basis$resp_levels))
+    out$Y <- as.integer(as_factor(out$Y, levels = x$basis$resp_levels))
   }
   if (is_ordinal(x$family) && is.ordered(out$Y)) {
-    out$Y <- as.numeric(out$Y)
+    diff <- ifelse(has_extra_cat(x$family), 1L, 0L)
+    out$Y <- as.integer(out$Y) - diff
   }
   if (check_response) {
     family4error <- family_names(x$family)
@@ -116,9 +122,12 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
       }
     }
     if (is_ordinal(x$family)) {
-      if (any(!is_wholenumber(out$Y)) || any(!out$Y > 0)) {
-        stop2("Family '", family4error, "' requires either positive ",
-              "integers or ordered factors as responses.")
+      extra_cat <- has_extra_cat(x$family)
+      min_int <- ifelse(extra_cat, 0L, 1L)
+      msg <- ifelse(extra_cat, "non-negative", "positive")
+      if (any(!is_wholenumber(out$Y)) || any(out$Y < min_int)) {
+        stop2("Family '", family4error, "' requires either ", msg,
+              " integers or ordered factors as responses.")
       }
     }
     if (use_int(x$family)) {
@@ -160,35 +169,23 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
     }
     out$Y <- as.array(out$Y)
   }
+
   # data for addition arguments of the response
+  # TODO: replace is.formula(x$adforms$term) pattern with has_ad_terms()
   if (has_trials(x$family) || is.formula(x$adforms$trials)) {
     if (!length(x$adforms$trials)) {
-      if (is_multinomial(x$family)) {
-        stop2("Specifying 'trials' is required in multinomial models.")
-      }
-      trials <- round(max(out$Y, na.rm = TRUE))
-      if (isTRUE(is.finite(trials))) {
-        message("Using the maximum response value as the number of trials.")
-        warning2(
-          "Using 'binomial' families without specifying 'trials' ",
-          "on the left-hand side of the model formula is deprecated."
-        )
-      } else if (!is.null(basis$trials)) {
-        trials <- max(basis$trials)
-      } else {
-        stop2("Could not compute the number of trials.")
-      }
-    } else if (is.formula(x$adforms$trials)) {
-      trials <- get_ad_values(x, "trials", "trials", data)
-      if (!is.numeric(trials)) {
-        stop2("Number of trials must be numeric.")
-      }
-      if (any(!is_wholenumber(trials) | trials < 0)) {
-        stop2("Number of trials must be non-negative integers.")
-      }
-    } else {
+      stop2("Specifying 'trials' is required for this model.")
+    }
+    if (!is.formula(x$adforms$trials)) {
       stop2("Argument 'trials' is misspecified.")
     }
+    trials <- get_ad_values(x, "trials", "trials", data)
+    if (!is.numeric(trials)) {
+      stop2("Number of trials must be numeric.")
+    }
+    if (any(!is_wholenumber(trials) | trials < 0)) {
+      stop2("Number of trials must be non-negative integers.")
+    }
     if (length(trials) == 1L) {
       trials <- rep(trials, nrow(data))
     }
@@ -322,16 +319,24 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
         "and refer to 'right' and 'none' respectively."
       )
     }
+    if (length(cens) == 1L) {
+      cens <- rep(cens, N)
+    }
+    if (length(cens) != N) {
+      stop2("Censoring information needs to have length ",
+            "equal to the number of data rows.")
+    }
     out$cens <- as.array(cens)
     icens <- cens %in% 2
-    y2_expr <- get_ad_expr(x, "cens", "y2")
-    if (any(icens) || !is.null(y2_expr)) {
+    if (any(icens) || has_interval_cens(x)) {
       # interval censoring is required
-      # check for 'y2' above as well to prevent issue #1367
       y2 <- unname(get_ad_values(x, "cens", "y2", data))
       if (is.null(y2)) {
         stop2("Argument 'y2' is required for interval censored data.")
       }
+      if (length(y2) != N) {
+        stop2("Argument 'y2' needs to have length equal to the number of data rows.")
+      }
       if (anyNA(y2[icens])) {
         stop2("'y2' should not be NA for interval censored observations.")
       }
@@ -437,19 +442,19 @@ data_response.brmsterms <- function(x, data, check_response = TRUE,
 }
 
 # data specific for mixture models
-data_mixture <- function(bterms, data2, prior) {
-  stopifnot(is.brmsterms(bterms))
+data_mixture <- function(bframe, data2, prior) {
+  stopifnot(is.brmsterms(bframe))
   out <- list()
-  if (is.mixfamily(bterms$family)) {
-    families <- family_names(bterms$family)
-    dp_classes <- dpar_class(names(c(bterms$dpars, bterms$fdpars)))
+  if (is.mixfamily(bframe$family)) {
+    families <- family_names(bframe$family)
+    dp_classes <- dpar_class(names(c(bframe$dpars, bframe$fdpars)))
     if (!any(dp_classes %in% "theta")) {
       # estimate mixture probabilities directly
-      take <- find_rows(prior, class = "theta", resp = bterms$resp)
+      take <- find_rows(prior, class = "theta", resp = bframe$resp)
       theta_prior <- prior$prior[take]
       con_theta <- eval_dirichlet(theta_prior, length(families), data2)
       out$con_theta <- as.array(con_theta)
-      p <- usc(combine_prefix(bterms))
+      p <- usc(combine_prefix(bframe))
       names(out) <- paste0(names(out), p)
     }
   }
@@ -457,20 +462,40 @@ data_mixture <- function(bterms, data2, prior) {
 }
 
 # data for the baseline functions of Cox models
-data_bhaz <- function(bterms, data, data2, prior, basis = NULL) {
+data_bhaz <- function(bframe, data, data2, prior) {
   out <- list()
-  if (!is_cox(bterms$family)) {
+  if (!is_cox(bframe$family)) {
     return(out)
   }
-  y <- model.response(model.frame(bterms$respform, data, na.action = na.pass))
-  args <- bterms$family$bhaz
-  bs <- basis$basis_matrix
-  out$Zbhaz <- bhaz_basis_matrix(y, args, basis = bs)
-  out$Zcbhaz <- bhaz_basis_matrix(y, args, integrate = TRUE, basis = bs)
+  y <- bframe$frame$resp$values
+  bhaz <- family_info(bframe, "bhaz")
+  bs <- bframe$basis$bhaz$basis_matrix
+  out$Zbhaz <- bhaz_basis_matrix(y, bhaz$args, basis = bs)
+  out$Zcbhaz <- bhaz_basis_matrix(y, bhaz$args, integrate = TRUE, basis = bs)
   out$Kbhaz <- NCOL(out$Zbhaz)
-  sbhaz_prior <- subset2(prior, class = "sbhaz", resp = bterms$resp)
-  con_sbhaz <- eval_dirichlet(sbhaz_prior$prior, out$Kbhaz, data2)
-  out$con_sbhaz <- as.array(con_sbhaz)
+  groups <- bhaz$groups
+  if (!is.null(groups)) {
+    out$ngrbhaz <- length(groups)
+    gr <- get_ad_values(bframe, "bhaz", "gr", data)
+    gr <- factor(rename(gr), levels = groups)
+    out$Jgrbhaz <- match(gr, groups)
+    out$con_sbhaz <- matrix(nrow = out$ngrbhaz, ncol = out$Kbhaz)
+    sbhaz_prior <- subset2(prior, class = "sbhaz", resp = bframe$resp)
+    sbhaz_prior_global <- subset2(sbhaz_prior, group = "")
+    con_sbhaz_global <- eval_dirichlet(sbhaz_prior_global$prior, out$Kbhaz, data2)
+    for (k in seq_along(groups)) {
+      sbhaz_prior_group <- subset2(sbhaz_prior, group = groups[k])
+      if (nzchar(sbhaz_prior_group$prior)) {
+        out$con_sbhaz[k, ] <- eval_dirichlet(sbhaz_prior_group$prior, out$Kbhaz, data2)
+      } else {
+        out$con_sbhaz[k, ] <- con_sbhaz_global
+      }
+    }
+  } else {
+    sbhaz_prior <- subset2(prior, class = "sbhaz", resp = bframe$resp)
+    con_sbhaz <- eval_dirichlet(sbhaz_prior$prior, out$Kbhaz, data2)
+    out$con_sbhaz <- as.array(con_sbhaz)
+  }
   out
 }
 
@@ -482,22 +507,20 @@ data_bhaz <- function(bterms, data, data2, prior, basis = NULL) {
 # @return the design matrix of the baseline hazard function
 bhaz_basis_matrix <- function(y, args = list(), integrate = FALSE,
                               basis = NULL) {
-  require_package("splines2")
+  # version check is required due to class name changes #1580
+  require_package("splines2", version = "0.5.0")
   if (!is.null(basis)) {
     # perform predictions based on an existing basis matrix
-    stopifnot(inherits(basis, "mSpline"))
+    stopifnot(inherits(basis, "MSpline"))
     if (integrate) {
-      # for predictions just the attibutes are required
+      # for predictions just the attributes are required
       # which are the same of M-Splines and I-Splines
-      class(basis) <- c("matrix", "iSpline")
+      class(basis) <- c("matrix", "ISpline")
     }
     return(predict(basis, y))
   }
   stopifnot(is.list(args))
   args$x <- y
-  if (!is.null(args$intercept)) {
-    args$intercept <- as_one_logical(args$intercept)
-  }
   if (is.null(args$Boundary.knots)) {
     # avoid 'knots' outside 'Boundary.knots' error (#1143)
     # we also need a smaller lower boundary knot to avoid lp = -Inf
@@ -517,6 +540,29 @@ bhaz_basis_matrix <- function(y, args = list(), integrate = FALSE,
   out
 }
 
+# extract baseline hazard information from data for storage in the model family
+# @return a named list with elements:
+#  args: arguments that can be passed to bhaz_basis_matrix
+#  groups: optional names of the groups for which to stratify
+extract_bhaz <- function(x, data) {
+  stopifnot(is.brmsformula(x) || is.brmsterms(x), is_cox(x))
+  if (is.null(x$adforms)) {
+    x$adforms <- terms_ad(x$formula, x$family)
+  }
+  out <- list()
+  if (is.null(x$adforms$bhaz)) {
+    # bhaz is an optional addition term so defaults need to be listed here too
+    out$args <- list(df = 5, intercept = TRUE)
+  } else {
+    out$args <- eval_rhs(x$adforms$bhaz)$flags
+    gr <- get_ad_values(x, "bhaz", "gr", data)
+    if (!is.null(gr)) {
+      out$groups <- rename(levels(factor(gr)))
+    }
+  }
+  out
+}
+
 # extract names of response categories
 # @param x a brmsterms object or one that can be coerced to it
 # @param data user specified data
@@ -543,7 +589,6 @@ extract_cat_names <- function(x, data) {
 # @return a data.frame with columns 'thres' and 'group'
 extract_thres_names <- function(x, data) {
   stopifnot(is.brmsformula(x) || is.brmsterms(x), has_thres(x))
-
   if (is.null(x$adforms)) {
     x$adforms <- terms_ad(x$formula, x$family)
   }
@@ -551,6 +596,8 @@ extract_thres_names <- function(x, data) {
   if (any(!is_wholenumber(nthres) | nthres < 1L)) {
     stop2("Number of thresholds must be a positive integer.")
   }
+  # has an extra category that is not part of the ordinal scale? (#1429)
+  extra_cat <- has_extra_cat(x$family)
   grthres <- get_ad_values(x, "thres", "gr", data)
   if (!is.null(grthres)) {
     # grouping variable was specified
@@ -564,7 +611,10 @@ extract_thres_names <- function(x, data) {
       nthres <- rep(NA, length(group))
       for (i in seq_along(group)) {
         take <- grthres %in% group[i]
-        nthres[i] <- extract_nthres(x$formula, data[take, , drop = FALSE])
+        nthres[i] <- extract_nthres(
+          x$formula, data[take, , drop = FALSE],
+          extra_cat = extra_cat
+        )
       }
     } else if (length(nthres) == 1L) {
       # replicate number of thresholds across groups
@@ -587,7 +637,7 @@ extract_thres_names <- function(x, data) {
     group <- ""
     if (!length(nthres)) {
       # extract number of thresholds from the response values
-      nthres <- extract_nthres(x$formula, data)
+      nthres <- extract_nthres(x$formula, data, extra_cat = extra_cat)
     }
     if (length(nthres) > 1L) {
       stop2("Number of thresholds needs to be a single value.")
@@ -597,17 +647,27 @@ extract_thres_names <- function(x, data) {
   data.frame(thres, group, stringsAsFactors = FALSE)
 }
 
-# extract threshold names from the response values
+# extract number of thresholds from the response values
 # @param formula with the response on the LHS
 # @param data a data.frame from which to extract responses
+# @param extra_cat is the first category an extra (hurdle) category?
 # @return a single value for the number of thresholds
-extract_nthres <- function(formula, data) {
+extract_nthres <- function(formula, data, extra_cat = FALSE) {
+  extra_cat <- as_one_logical(extra_cat)
   respform <- validate_resp_formula(formula)
   mr <- model.response(model.frame(respform, data))
   if (is_like_factor(mr)) {
-    out <- length(levels(factor(mr))) - 1
+    # the first factor level is the extra category
+    diff <- ifelse(extra_cat, 2L, 1L)
+    out <- length(levels(factor(mr))) - diff
   } else {
-    out <- max(mr) - 1
+    # 0 is the extra category which does not affect max
+    out <- max(mr) - 1L
+  }
+  if (out < 1L) {
+    stop2("Could not extract the number of thresholds. Use ordered factors ",
+          "or positive integers as your ordinal response and ensure that ",
+          "more than on response category is present.")
   }
   out
 }
