@@ -47,11 +47,12 @@ pstudent_t <- function(q, df, mu = 0, sigma = 1,
 
 #' @rdname StudentT
 #' @export
-qstudent_t <-  function(p, df, mu = 0, sigma = 1) {
+qstudent_t <-  function(p, df, mu = 0, sigma = 1,
+                        lower.tail = TRUE, log.p = FALSE) {
   if (isTRUE(any(sigma < 0))) {
     stop2("sigma must be non-negative.")
   }
-  mu + sigma * qt(p, df = df)
+  mu + sigma * qt(p, df = df, lower.tail = lower.tail, log.p = log.p)
 }
 
 #' @rdname StudentT
@@ -125,7 +126,7 @@ rmulti_normal <- function(n, mu, Sigma, check = FALSE) {
     }
   }
   draws <- matrix(rnorm(n * p), nrow = n, ncol = p)
-  mu + draws %*% chol(Sigma)
+  sweep(draws %*% chol(Sigma), 2, mu, "+")
 }
 
 #' The Multivariate Student-t Distribution
@@ -299,7 +300,13 @@ pskew_normal <- function(q, mu = 0, sigma = 1, alpha = 0,
     delta[is_alpha_inf] <- sign(alpha[is_alpha_inf])
     out <- numeric(nz)
     for (k in seq_len(nz)) {
-      if (is_alpha_inf[k]) {
+      if (is.infinite(z[k])) {
+        if (z[k] > 0) {
+          out[k] <- 1
+        } else {
+          out[k] <- 0
+        }
+      } else if (is_alpha_inf[k]) {
         if (alpha[k] > 0) {
           out[k] <- 2 * (pnorm(pmax(z[k], 0)) - 0.5)
         } else {
@@ -333,12 +340,7 @@ qskew_normal <- function(p, mu = 0, sigma = 1, alpha = 0,
   if (isTRUE(any(sigma < 0))) {
     stop2("sigma must be non-negative.")
   }
-  if (log.p) {
-    p <- exp(p)
-  }
-  if (!lower.tail) {
-    p <- 1 - p
-  }
+  p <- validate_p_dist(p, lower.tail = lower.tail, log.p = log.p)
   args <- cp2dp(mu, sigma, alpha, xi = xi, omega = omega, p = p)
   out <- with(args, {
     # do it like sn::qsn
@@ -459,7 +461,7 @@ pinvgamma <- function(q, shape, rate, lower.tail = TRUE, log.p = FALSE) {
 #' @name VonMises
 #'
 #' @inheritParams StudentT
-#' @param x,q Vector of quantiles.
+#' @param x,q Vector of quantiles between \code{-pi} and \code{pi}.
 #' @param kappa Vector of precision values.
 #' @param acc Accuracy of numerical approximations.
 #'
@@ -752,21 +754,13 @@ pfrechet <- function(q, loc = 0, scale = 1, shape = 1,
 #' @export
 qfrechet <- function(p, loc = 0, scale = 1, shape = 1,
                      lower.tail = TRUE, log.p = FALSE) {
-  if (isTRUE(any(p <= 0)) || isTRUE(any(p >= 1))) {
-    stop("'p' must contain probabilities in (0,1)")
-  }
   if (isTRUE(any(scale <= 0))) {
     stop2("Argument 'scale' must be positive.")
   }
   if (isTRUE(any(shape <= 0))) {
     stop2("Argument 'shape' must be positive.")
   }
-  if (log.p) {
-    p <- exp(p)
-  }
-  if (!lower.tail) {
-    p <- 1 - p
-  }
+  p <- validate_p_dist(p, lower.tail = lower.tail, log.p = log.p)
   loc + scale * (-log(p))^(-1/shape)
 }
 
@@ -1020,6 +1014,24 @@ pgen_extreme_value <- function(q, mu = 0, sigma = 1, xi = 0,
   out
 }
 
+#' @rdname GenExtremeValue
+#' @export
+qgen_extreme_value <- function(p, mu = 0, sigma = 1, xi = 0,
+                               lower.tail = TRUE, log.p = FALSE) {
+  if (isTRUE(any(sigma <= 0))) {
+    stop2("sigma bust be positive.")
+  }
+  p <- validate_p_dist(p, lower.tail = lower.tail, log.p = log.p)
+  args <- nlist(p, mu, sigma, xi)
+  args <- do_call(expand, args)
+  out <- with(args, ifelse(
+    xi == 0,
+    mu - sigma * log(-log(p)),
+    mu - sigma * (1 - (-log(p)) ^ (-xi)) / xi
+  ))
+  out
+}
+
 #' @rdname GenExtremeValue
 #' @export
 rgen_extreme_value <- function(n, mu = 0, sigma = 1, xi = 0) {
@@ -1089,12 +1101,7 @@ pasym_laplace <- function(q, mu = 0, sigma = 1, quantile = 0.5,
 #' @export
 qasym_laplace <- function(p, mu = 0, sigma = 1, quantile = 0.5,
                           lower.tail = TRUE, log.p = FALSE) {
-  if (log.p) {
-    p <- exp(p)
-  }
-  if (!lower.tail) {
-    p <- 1 - p
-  }
+  p <- validate_p_dist(p, lower.tail = lower.tail, log.p = log.p)
   if (length(quantile) == 1L) {
     quantile <- rep(quantile, length(mu))
   }
@@ -1175,12 +1182,7 @@ qdiscrete_weibull <- function(p, mu, shape, lower.tail = TRUE, log.p = FALSE) {
   if (isTRUE(any(shape <= 0))) {
     stop2("shape bust be positive.")
   }
-  if (log.p) {
-    p <- exp(p)
-  }
-  if (!lower.tail) {
-    p <- 1 - p
-  }
+  p <- validate_p_dist(p, lower.tail = lower.tail, log.p = log.p)
   ceiling((log(1 - p) / log(mu))^(1 / shape) - 1)
 }
 
@@ -2541,6 +2543,21 @@ rshifted <- function(dist, n, shift = 0, ...) {
   do_call(paste0("r", dist), list(n, ...)) + shift
 }
 
+# validate argument p in q<dist> functions
+validate_p_dist <- function(p, lower.tail = TRUE, log.p = FALSE) {
+  if (log.p) {
+    p <- exp(p)
+  }
+  if (!lower.tail) {
+    p <- 1 - p
+  }
+  if (isTRUE(any(p < 0)) || isTRUE(any(p > 1))) {
+    p[p < 0 | p > 1] <- NaN
+    warning2("NaNs produced")
+  }
+  p
+}
+
 # check if 'n' in r<dist> functions is valid
 # @param n number of desired random draws
 # @param .. parameter vectors
