@@ -2,11 +2,11 @@
 #'
 #' Specify autocorrelation terms in \pkg{brms} models. Currently supported terms
 #' are \code{\link{arma}}, \code{\link{ar}}, \code{\link{ma}},
-#' \code{\link{cosy}}, \code{\link{sar}}, \code{\link{car}}, and
-#' \code{\link{fcor}}. Terms can be directly specified within the formula, or
-#' passed to the \code{autocor} argument of \code{\link{brmsformula}} in the
-#' form of a one-sided formula. For deprecated ways of specifying
-#' autocorrelation terms, see \code{\link{cor_brms}}.
+#' \code{\link{cosy}}, \code{\link{unstr}}, \code{\link{sar}},
+#' \code{\link{car}}, and \code{\link{fcor}}. Terms can be directly specified
+#' within the formula, or passed to the \code{autocor} argument of
+#' \code{\link{brmsformula}} in the form of a one-sided formula. For deprecated
+#' ways of specifying autocorrelation terms, see \code{\link{cor_brms}}.
 #'
 #' @name autocor-terms
 #'
@@ -17,8 +17,8 @@
 #'
 #' @seealso \code{\link{brmsformula}}, \code{\link{acformula}},
 #'   \code{\link{arma}}, \code{\link{ar}}, \code{\link{ma}},
-#'   \code{\link{cosy}}, \code{\link{sar}}, \code{\link{car}},
-#'   \code{\link{fcor}}
+#'   \code{\link{cosy}}, \code{\link{unstr}}, \code{\link{sar}},
+#'   \code{\link{car}}, \code{\link{fcor}}
 #'
 #' @examples
 #' # specify autocor terms within the formula
@@ -72,9 +72,9 @@ NULL
 #'
 #' @export
 arma <- function(time = NA, gr = NA, p = 1, q = 1, cov = FALSE) {
-  label <- deparse(match.call())
-  time <- deparse(substitute(time))
-  gr <- deparse(substitute(gr))
+  label <- deparse0(match.call())
+  time <- deparse0(substitute(time))
+  gr <- deparse0(substitute(gr))
   .arma(time = time, gr = gr, p = p, q = q, cov = cov, label = label)
 }
 
@@ -102,9 +102,9 @@ arma <- function(time = NA, gr = NA, p = 1, q = 1, cov = FALSE) {
 #'
 #' @export
 ar <- function(time = NA, gr = NA, p = 1, cov = FALSE) {
-  label <- deparse(match.call())
-  time <- deparse(substitute(time))
-  gr <- deparse(substitute(gr))
+  label <- deparse0(match.call())
+  time <- deparse0(substitute(time))
+  gr <- deparse0(substitute(gr))
   .arma(time = time, gr = gr, p = p, q = 0, cov = cov, label = label)
 }
 
@@ -132,9 +132,9 @@ ar <- function(time = NA, gr = NA, p = 1, cov = FALSE) {
 #'
 #' @export
 ma <- function(time = NA, gr = NA, q = 1, cov = FALSE) {
-  label <- deparse(match.call())
-  time <- deparse(substitute(time))
-  gr <- deparse(substitute(gr))
+  label <- deparse0(match.call())
+  time <- deparse0(substitute(time))
+  gr <- deparse0(substitute(gr))
   .arma(time = time, gr = gr, p = 0, q = q, cov = cov, label = label)
 }
 
@@ -188,18 +188,50 @@ ma <- function(time = NA, gr = NA, q = 1, cov = FALSE) {
 #' }
 #'
 #' @export
-#' @export
 cosy <- function(time = NA, gr = NA) {
-  label <- deparse(match.call())
-  time <- deparse(substitute(time))
+  label <- deparse0(match.call())
+  time <- deparse0(substitute(time))
   time <- as_one_variable(time)
-  gr <- deparse(substitute(gr))
+  gr <- deparse0(substitute(gr))
   stopif_illegal_group(gr)
   out <- nlist(time, gr, label)
   class(out) <- c("cosy_term", "ac_term")
   out
 }
 
+#' Set up UNSTR correlation structures
+#'
+#' Set up an unstructured (UNSTR) correlation term in \pkg{brms}. The function does
+#' not evaluate its arguments -- it exists purely to help set up a model with
+#' UNSTR terms.
+#'
+#' @inheritParams arma
+#'
+#' @return An object of class \code{'unstr_term'}, which is a list
+#'   of arguments to be interpreted by the formula
+#'   parsing functions of \pkg{brms}.
+#'
+#' @seealso \code{\link{autocor-terms}}
+#'
+#' @examples
+#' \dontrun{
+#' # add an unstructured correlation matrix for visits within the same patient
+#' fit <- brm(count ~ Trt + unstr(visit, patient), data = epilepsy)
+#' summary(fit)
+#' }
+#'
+#' @export
+unstr <- function(time, gr) {
+  label <- deparse0(match.call())
+  time <- deparse0(substitute(time))
+  time <- as_one_variable(time)
+  gr <- deparse0(substitute(gr))
+  stopif_illegal_group(gr)
+  out <- nlist(time, gr, label)
+  class(out) <- c("unstr_term", "ac_term")
+  out
+}
+
 #' Spatial simultaneous autoregressive (SAR) structures
 #'
 #' Set up an spatial simultaneous autoregressive (SAR) term in \pkg{brms}. The
@@ -247,11 +279,11 @@ cosy <- function(time = NA, gr = NA) {
 #'
 #' @export
 sar <- function(M, type = "lag") {
-  label <- deparse(match.call())
+  label <- deparse0(match.call())
   if (missing(M)) {
     stop2("Argument 'M' is missing in sar().")
   }
-  M <- deparse(substitute(M))
+  M <- deparse0(substitute(M))
   M <- as_one_variable(M)
   options <- c("lag", "error")
   type <- match.arg(type, options)
@@ -301,6 +333,7 @@ sar <- function(M, type = "lag") {
 #' distance <- as.matrix(dist(Grid))
 #' W <- array(0, c(K, K))
 #' W[distance == 1] <- 1
+#' rownames(W) <- 1:nrow(W)
 #'
 #' # generate the covariates and response data
 #' x1 <- rnorm(K)
@@ -313,10 +346,11 @@ sar <- function(M, type = "lag") {
 #' prob <- exp(eta) / (1 + exp(eta))
 #' size <- rep(50, K)
 #' y <- rbinom(n = K, size = size, prob = prob)
-#' dat <- data.frame(y, size, x1, x2)
+#' g <- 1:length(y)
+#' dat <- data.frame(y, size, x1, x2, g)
 #'
 #' # fit a CAR model
-#' fit <- brm(y | trials(size) ~ x1 + x2 + car(W),
+#' fit <- brm(y | trials(size) ~ x1 + x2 + car(W, gr = g),
 #'            data = dat, data2 = list(W = W),
 #'            family = binomial())
 #' summary(fit)
@@ -324,13 +358,13 @@ sar <- function(M, type = "lag") {
 #'
 #' @export
 car <- function(M, gr = NA, type = "escar") {
-  label <- deparse(match.call())
+  label <- deparse0(match.call())
   if (missing(M)) {
     stop2("Argument 'M' is missing in car().")
   }
-  M <- deparse(substitute(M))
+  M <- deparse0(substitute(M))
   M <- as_one_variable(M)
-  gr <- deparse(substitute(gr))
+  gr <- deparse0(substitute(gr))
   stopif_illegal_group(gr)
   options <- c("escar", "esicar", "icar", "bym2")
   type <- match.arg(type, options)
@@ -367,11 +401,11 @@ car <- function(M, gr = NA, type = "escar") {
 #'
 #' @export
 fcor <- function(M) {
-  label <- deparse(match.call())
+  label <- deparse0(match.call())
   if (missing(M)) {
     stop2("Argument 'M' is missing in fcor().")
   }
-  M <- deparse(substitute(M))
+  M <- deparse0(substitute(M))
   M <- as_one_variable(M)
   out <- nlist(M, label)
   class(out) <- c("fcor_term", "ac_term")
@@ -403,31 +437,31 @@ validate_autocor <- function(autocor) {
 
 # gather information on autocor terms
 # @return a data.frame with one row per autocor term
-tidy_acef <- function(x, ...) {
-  UseMethod("tidy_acef")
+frame_ac <- function(x, ...) {
+  UseMethod("frame_ac")
 }
 
 #' @export
-tidy_acef.default <- function(x, ...) {
+frame_ac.default <- function(x, ...) {
   x <- brmsterms(x, check_response = FALSE)
-  tidy_acef(x, ...)
+  frame_ac(x, ...)
 }
 
 #' @export
-tidy_acef.mvbrmsterms <- function(x, ...) {
-  out <- lapply(x$terms, tidy_acef, ...)
+frame_ac.mvbrmsterms <- function(x, ...) {
+  out <- lapply(x$terms, frame_ac, ...)
   out <- do_call(rbind, out)
-  structure(out, class = acef_class())
+  structure(out, class = acframe_class())
 }
 
 #' @export
-tidy_acef.brmsterms <- function(x, ...) {
-  out <- lapply(x$dpars, tidy_acef, ...)
+frame_ac.brmsterms <- function(x, ...) {
+  out <- lapply(x$dpars, frame_ac, ...)
   out <- do_call(rbind, out)
   if (!NROW(out)) {
-    return(empty_acef())
+    return(empty_acframe())
   }
-  out <- structure(out, class = acef_class())
+  out <- structure(out, class = acframe_class())
   if (has_ac_class(out, "sar")) {
     if (any(c("sigma", "nu") %in% names(x$dpars))) {
       stop2("SAR models are not implemented when predicting 'sigma' or 'nu'.")
@@ -443,10 +477,10 @@ tidy_acef.brmsterms <- function(x, ...) {
 }
 
 #' @export
-tidy_acef.btl <- function(x, data = NULL, ...) {
+frame_ac.btl <- function(x, data = NULL, ...) {
   form <- x[["ac"]]
   if (!is.formula(form)) {
-    return(empty_acef())
+    return(empty_acframe())
   }
   if (is.mixfamily(x$family)) {
     stop2("Autocorrelation terms cannot be applied in mixture models.")
@@ -457,6 +491,7 @@ tidy_acef.btl <- function(x, data = NULL, ...) {
   cnames <- c("class", "dim", "type", "time", "gr", "p", "q", "M")
   out[cnames] <- list(NA)
   out$cov <- out$nat_cov <- FALSE
+  out$nat_res <- has_natural_residuals(x)
   out[names(px)] <- px
   for (i in seq_len(nterms)) {
     ac <- eval2(out$term[i])
@@ -476,6 +511,13 @@ tidy_acef.btl <- function(x, data = NULL, ...) {
       out$gr[i] <- ac$gr
       out$cov[i] <- TRUE
     }
+    if (is.unstr_term(ac)) {
+      out$class[i] <- "unstr"
+      out$dim[i] <- "time"
+      out$time[i] <- ac$time
+      out$gr[i] <- ac$gr
+      out$cov[i] <- TRUE
+    }
     if (is.sar_term(ac)) {
       out$class[i] <- "sar"
       out$dim[i] <- "space"
@@ -499,8 +541,8 @@ tidy_acef.btl <- function(x, data = NULL, ...) {
   # covariance matrices of natural residuals will be handled
   # directly in the likelihood function while latent residuals will
   # be added to the linear predictor of the main parameter 'mu'
-  out$nat_cov <- out$cov & has_natural_residuals(x)
-  class(out) <- acef_class()
+  out$nat_cov <- out$cov & out$nat_res
+  class(out) <- acframe_class()
   # validate specified autocor terms
   if (any(duplicated(out$class))) {
     stop2("Can only model one term per autocorrelation class.")
@@ -519,37 +561,49 @@ tidy_acef.btl <- function(x, data = NULL, ...) {
       stop2("Explicit covariance terms can only be specified on 'mu'.")
     }
   }
+  if (!is.null(data)) {
+    # optional such that this function can be applied
+    # without data before brmsframe is being created
+    time <- get_ac_vars(out, "time", dim = "time")
+    if (length(time)) {
+      attr(out, "times") <- extract_levels(get(time, data))
+    }
+  }
   out
 }
 
 #' @export
-tidy_acef.btnl <- function(x, ... ) {
-  tidy_acef.btl(x, ...)
+frame_ac.btnl <- function(x, ... ) {
+  frame_ac.btl(x, ...)
 }
 
 #' @export
-tidy_acef.acef <- function(x, ...) {
+frame_ac.acframe <- function(x, ...) {
   x
 }
 
 #' @export
-tidy_acef.NULL <- function(x, ...) {
-  empty_acef()
+frame_ac.NULL <- function(x, ...) {
+  empty_acframe()
 }
 
-empty_acef <- function() {
-  structure(empty_data_frame(), class = acef_class())
+empty_acframe <- function() {
+  structure(empty_data_frame(), class = acframe_class())
 }
 
-acef_class <- function() {
-  c("acef", "data.frame")
+acframe_class <- function() {
+  c("acframe", "data.frame")
+}
+
+is.acframe <- function(x) {
+  inherits(x, "acframe")
 }
 
 # get names of certain autocor variables
 get_ac_vars <- function(x, var, ...) {
   var <- match.arg(var, c("time", "gr", "M"))
-  acef <- subset2(tidy_acef(x), ...)
-  out <- unique(acef[[var]])
+  acframe <- subset2(frame_ac(x), ...)
+  out <- unique(acframe[[var]])
   setdiff(na.omit(out), "NA")
 }
 
@@ -560,7 +614,7 @@ get_ac_groups <- function(x, ...) {
 
 # is certain subset of autocor terms is present?
 has_ac_subset <- function(x, ...) {
-  NROW(subset2(tidy_acef(x), ...)) > 0L
+  NROW(subset2(frame_ac(x), ...)) > 0L
 }
 
 # is a certain autocorrelation class present?
@@ -578,10 +632,16 @@ use_ac_cov_time <- function(x) {
   has_ac_subset(x, cov = TRUE, dim = "time")
 }
 
+# check if the family has natural residuals
+has_ac_natural_residuals <- function(x) {
+  has_ac_subset(x, nat_res = TRUE)
+}
+
 # does the model need latent residuals for autocor structures?
-has_ac_latent_residuals <- function(bterms) {
-  !has_natural_residuals(bterms) &&
-    (use_ac_cov(bterms) || has_ac_class(bterms, "arma"))
+has_ac_latent_residuals <- function(x) {
+  x <- frame_ac(x)
+  !has_ac_natural_residuals(x) &&
+    (use_ac_cov(x) || has_ac_class(x, "arma"))
 }
 
 # validate SAR matrices
@@ -638,6 +698,7 @@ validate_fcor_matrix <- function(M) {
 
 # regex to extract all parameter names of autocorrelation structures
 regex_autocor_pars <- function() {
+  # cortime is ignored here to allow custom renaming in summary.brmsfit
   p <- c("ar", "ma", "sderr", "cosy", "lagsar", "errorsar",
          "car", "sdcar", "rhocar")
   p <- paste0("(", p, ")", collapse = "|")
@@ -656,6 +717,10 @@ is.cosy_term <- function(x) {
   inherits(x, "cosy_term")
 }
 
+is.unstr_term <- function(x) {
+  inherits(x, "unstr_term")
+}
+
 is.sar_term <- function(x) {
   inherits(x, "sar_term")
 }
