@@ -43,6 +43,7 @@
 #' @param denom A vector of positive numeric values specifying
 #'   the denominator values from which the response rates are computed.
 #' @param gr A vector of grouping indicators.
+#' @param df Degrees of freedom of baseline hazard splines for Cox models.
 #' @param ... For \code{resp_vreal}, vectors of real values.
 #'   For \code{resp_vint}, vectors of integer values. In Stan,
 #'   these variables will be named \code{vreal1}, \code{vreal2}, ...,
@@ -106,7 +107,7 @@ NULL
 #' @rdname addition-terms
 #' @export
 resp_se <- function(x, sigma = FALSE) {
-  se <- deparse(substitute(x))
+  se <- deparse0(substitute(x))
   sigma <- as_one_logical(sigma)
   class_resp_special(
     "se", call = match.call(),
@@ -117,7 +118,7 @@ resp_se <- function(x, sigma = FALSE) {
 #' @rdname addition-terms
 #' @export
 resp_weights <- function(x, scale = FALSE) {
-  weights <- deparse(substitute(x))
+  weights <- deparse0(substitute(x))
   scale <- as_one_logical(scale)
   class_resp_special(
     "weights", call = match.call(),
@@ -128,15 +129,15 @@ resp_weights <- function(x, scale = FALSE) {
 #' @rdname addition-terms
 #' @export
 resp_trials <- function(x) {
-  trials <- deparse(substitute(x))
+  trials <- deparse0(substitute(x))
   class_resp_special("trials", call = match.call(), vars = nlist(trials))
 }
 
 #' @rdname addition-terms
 #' @export
 resp_thres <- function(x, gr = NA) {
-  thres <- deparse(substitute(x))
-  gr <- deparse(substitute(gr))
+  thres <- deparse0(substitute(x))
+  gr <- deparse0(substitute(gr))
   class_resp_special("thres", call = match.call(), vars = nlist(thres, gr))
 }
 
@@ -145,7 +146,7 @@ resp_thres <- function(x, gr = NA) {
 resp_cat <- function(x) {
   # deprecated as of brms 2.10.5
   # number of thresholds = number of response categories - 1
-  thres <- deparse(substitute(x))
+  thres <- deparse0(substitute(x))
   str_add(thres) <- " - 1"
   class_resp_special(
     "thres", call = match.call(),
@@ -156,51 +157,62 @@ resp_cat <- function(x) {
 #' @rdname addition-terms
 #' @export
 resp_dec <- function(x) {
-  dec <- deparse(substitute(x))
+  dec <- deparse0(substitute(x))
   class_resp_special("dec", call = match.call(), vars = nlist(dec))
 }
 
+#' @rdname addition-terms
+#' @export
+resp_bhaz <- function(gr = NA, df = 5, ...) {
+  gr <- deparse0(substitute(gr))
+  df <- as_one_integer(df)
+  args <- nlist(df, ...)
+  # non-power users shouldn't know they can change 'intercept'
+  args$intercept <- args$intercept %||% TRUE
+  class_resp_special("bhaz", call = match.call(), vars = nlist(gr), flags = args)
+}
+
 #' @rdname addition-terms
 #' @export
 resp_cens <- function(x, y2 = NA) {
-  cens <- deparse(substitute(x))
-  y2 <- deparse(substitute(y2))
+  cens <- deparse0(substitute(x))
+  y2 <- deparse0(substitute(y2))
   class_resp_special("cens", call = match.call(), vars = nlist(cens, y2))
 }
 
 #' @rdname addition-terms
 #' @export
 resp_trunc <- function(lb = -Inf, ub = Inf) {
-  lb <- deparse(substitute(lb))
-  ub <- deparse(substitute(ub))
+  lb <- deparse0(substitute(lb))
+  ub <- deparse0(substitute(ub))
   class_resp_special("trunc", call = match.call(), vars = nlist(lb, ub))
 }
 
 #' @rdname addition-terms
 #' @export
 resp_mi <- function(sdy = NA) {
-  sdy <- deparse(substitute(sdy))
+  sdy <- deparse0(substitute(sdy))
   class_resp_special("mi", call = match.call(), vars = nlist(sdy))
 }
 
 #' @rdname addition-terms
 #' @export
 resp_index <- function(x) {
-  index <- deparse(substitute(x))
+  index <- deparse0(substitute(x))
   class_resp_special("index", call = match.call(), vars = nlist(index))
 }
 
 #' @rdname addition-terms
 #' @export
 resp_rate <- function(denom) {
-  denom <- deparse(substitute(denom))
+  denom <- deparse0(substitute(denom))
   class_resp_special("rate", call = match.call(), vars = nlist(denom))
 }
 
 #' @rdname addition-terms
 #' @export
 resp_subset <- function(x) {
-  subset <- deparse(substitute(x))
+  subset <- deparse0(substitute(x))
   class_resp_special("subset", call = match.call(), vars = nlist(subset))
 }
 
@@ -226,7 +238,7 @@ resp_vint <- function(...) {
 class_resp_special <- function(type, call, vars = list(), flags = list()) {
   type <- as_one_character(type)
   stopifnot(is.call(call), is.list(vars), is.list(flags))
-  label <- deparse(call)
+  label <- deparse0(call)
   out <- nlist(type, call, label, vars, flags)
   class(out) <- c("resp_special")
   out
@@ -327,6 +339,11 @@ get_cens <- function(bterms, data, resp = NULL) {
   out
 }
 
+# indicates if the model may have interval censored observations
+has_interval_cens <- function(bterms) {
+  !is.null(get_ad_expr(bterms, "cens", "y2"))
+}
+
 # extract truncation boundaries
 # @param bterms a brmsterms object
 # @param data data.frame containing the truncation variables
@@ -364,24 +381,28 @@ trunc_bounds <- function(bterms, data = NULL, incl_family = FALSE,
   out
 }
 
-# check if addition argument 'subset' ist used in the model
+# check if addition argument 'subset' is used in the model
+# works for both univariate and multivariate models
 has_subset <- function(bterms) {
-  .has_subset <- function(x) {
-    is.formula(x$adforms$subset)
-  }
   if (is.brmsterms(bterms)) {
-    out <- .has_subset(bterms)
+    out <- has_ad_terms(bterms, "subset")
   } else if (is.mvbrmsterms(bterms)) {
-    out <- any(ulapply(bterms$terms, .has_subset))
+    out <- any(ulapply(bterms$terms, has_ad_terms, "subset"))
   } else {
     out <- FALSE
   }
   out
 }
 
+# check if a model has certain addition terms
+has_ad_terms <- function(bterms, terms) {
+  stopifnot(is.brmsterms(bterms), is.character(terms))
+  any(ulapply(bterms$adforms[terms], is.formula))
+}
+
 # construct a list of indices for cross-formula referencing
-tidy_index <- function(x, data) {
-  out <- .tidy_index(x, data)
+frame_index <- function(x, data) {
+  out <- .frame_index(x, data)
   if (is.brmsterms(x)) {
     # ensure consistent format for both uni- and multivariate models
     out <- list(out)
@@ -390,13 +411,13 @@ tidy_index <- function(x, data) {
   out
 }
 
-# internal version of tidy_index
-.tidy_index <- function(x, ...) {
-  UseMethod(".tidy_index")
+# internal version of frame_index
+.frame_index <- function(x, ...) {
+  UseMethod(".frame_index")
 }
 
 #' @export
-.tidy_index.brmsterms <- function(x, data, ...) {
+.frame_index.brmsterms <- function(x, data, ...) {
   out <- get_ad_values(x, "index", "index", data)
   if (is.null(out)) {
     return(NULL)
@@ -410,14 +431,14 @@ tidy_index <- function(x, data) {
     stop2("NAs are not allowed in 'index' variables.")
   }
   if (anyDuplicated(out)) {
-    stop2("Index of response '", names(out), "' contains duplicated values.")
+    stop2("Index of response '", x$resp, "' contains duplicated values.")
   }
   out
 }
 
 #' @export
-.tidy_index.mvbrmsterms <- function(x, data, ...) {
-  lapply(x$terms, .tidy_index, data = data, ...)
+.frame_index.mvbrmsterms <- function(x, data, ...) {
+  lapply(x$terms, .frame_index, data = data, ...)
 }
 
 # check if cross-formula referencing is possible in subsetted models
