@@ -27,16 +27,18 @@
 #'  details.
 #' @param reloo Logical; Indicate whether \code{\link{reloo}}
 #'  should be applied on problematic observations. Defaults to \code{FALSE}.
-#' @param k_threshold The threshold at which pareto \eqn{k}
-#'   estimates are treated as problematic. Defaults to \code{0.7}.
-#'   Only used if argument \code{reloo} is \code{TRUE}.
+#' @param k_threshold The Pareto \eqn{k} threshold for which observations
+#'   \code{\link{loo_moment_match}} or \code{\link{reloo}} is applied if
+#'   argument \code{moment_match} or \code{reloo} is \code{TRUE}.
+#'   Defaults to \code{0.7}.
 #'   See \code{\link[loo:pareto-k-diagnostic]{pareto_k_ids}} for more details.
 #' @param save_psis Should the \code{"psis"} object created internally be saved
 #'   in the returned object? For more details see \code{\link[loo:loo]{loo}}.
-#' @param moment_match_args Optional \code{list} of additional arguments passed to
-#'   \code{\link{loo_moment_match}}.
-#' @param reloo_args Optional \code{list} of additional arguments passed to
-#'   \code{\link{reloo}}.
+#' @param moment_match_args Optional named \code{list} of additional arguments
+#'  passed to \code{\link{loo_moment_match}}.
+#' @param reloo_args Optional named \code{list} of additional arguments passed to
+#'   \code{\link{reloo}}. This can be useful, among others, to control
+#'   how many chains, iterations, etc. to use for the fitted sub-models.
 #' @param model_names If \code{NULL} (the default) will use model names
 #'   derived from deparsing the call. Otherwise will use the passed
 #'   values as model names.
@@ -88,6 +90,14 @@ loo.brmsfit <-  function(x, ..., compare = TRUE, resp = NULL,
                          moment_match_args = list(), reloo_args = list(),
                          model_names = NULL) {
   args <- split_dots(x, ..., model_names = model_names)
+  if (!"use_stored" %in% names(args)) {
+    further_arg_names <- c(
+      "resp", "moment_match", "reloo", "k_threshold",
+      "save_psis", "moment_match_args", "reloo_args"
+    )
+    args$use_stored <- all(names(args) %in% "models") &&
+      !any(further_arg_names %in% names(match.call()))
+  }
   c(args) <- nlist(
     criterion = "loo", pointwise, compare,
     resp, k_threshold, save_psis, moment_match,
@@ -165,6 +175,11 @@ LOO <- function(x, ...) {
 waic.brmsfit <- function(x, ..., compare = TRUE, resp = NULL,
                          pointwise = FALSE, model_names = NULL) {
   args <- split_dots(x, ..., model_names = model_names)
+  if (!"use_stored" %in% names(args)) {
+    further_arg_names <- c("resp")
+    args$use_stored <- all(names(args) %in% "models") &&
+      !any(further_arg_names %in% names(match.call()))
+  }
   c(args) <- nlist(criterion = "waic", pointwise, compare, resp)
   do_call(compute_loolist, args)
 }
@@ -242,7 +257,10 @@ compute_loo <- function(x, criterion, newdata = NULL, resp = NULL,
   model_name <- as_one_character(model_name)
   use_stored <- as_one_logical(use_stored)
   out <- get_criterion(x, criterion)
-  if (!(use_stored && is.loo(out))) {
+  if (is.loo(out) && !use_stored) {
+    message("Recomputing '", criterion, "' for model '", model_name, "'")
+  }
+  if (!is.loo(out) || !use_stored) {
     args <- nlist(x, newdata, resp, model_name, ...)
     out <- do_call(paste0(".", criterion), args)
     attr(out, "yhash") <- hash_response(x, newdata = newdata, resp = resp)
@@ -294,11 +312,35 @@ loo_criteria <- function() {
   do_call("waic", loo_args, pkg = "loo")
 }
 
-# compute 'psis' criterion using the 'loo' package
-# @param model_name ignored but included to avoid being passed to '...'
+# alias of psis for convenient use in compute_loo()
 .psis <- function(x, newdata, resp, model_name, ...) {
+  psis(x, newdata = newdata, resp = resp, model_name = model_name, ...)
+}
+
+#' @inherit loo::psis return title description details references
+#'
+#' @aliases psis psis.brmsfit
+#'
+#' @param log_ratios A fitted model object of class \code{brmsfit}.
+#'   Argument is named "log_ratios" to match the argument name of the
+#'   \code{\link[loo:psis]{loo::psis}} generic function.
+#' @param model_name Currently ignored.
+#' @param ... Further arguments passed to \code{\link{log_lik}} and
+#'   \code{\link[loo:psis]{loo::psis}}.
+#' @inheritParams log_lik.brmsfit
+#'
+#' @examples
+#' \dontrun{
+#' fit <- brm(rating ~ treat + period + carry, data = inhaler)
+#' psis(fit)
+#'}
+#' @importFrom loo psis
+#' @export psis
+#' @export
+psis.brmsfit <- function(log_ratios, newdata = NULL, resp = NULL,
+                         model_name = NULL, ...) {
   loo_args <- prepare_loo_args(
-    x, newdata = newdata, resp = resp,
+    log_ratios, newdata = newdata, resp = resp,
     pointwise = FALSE, ...
   )
   loo_args$log_ratios <- -loo_args$x
@@ -365,13 +407,16 @@ loo_compare.brmsfit <- function(x, ..., criterion = c("loo", "waic", "kfold"),
   loos <- named_list(names(models))
   for (i in seq_along(models)) {
     models[[i]] <- restructure(models[[i]])
-    loos[[i]] <- get_criterion(models[[i]], criterion)
-    if (is.null(loos[[i]])) {
+    loo_i <- get_criterion(models[[i]], criterion)
+    if (is.null(loo_i)) {
       stop2(
         "Model '", names(models)[i], "' does not contain a precomputed '",
         criterion, "' criterion. See ?loo_compare.brmsfit for help."
       )
     }
+    # only assign object to list after checking if non-null
+    # otherwise the index may be out of bounds in the error check
+    loos[[i]] <- loo_i
   }
   loo_compare(loos)
 }
@@ -433,6 +478,8 @@ loo_model_weights.brmsfit <- function(x, ..., model_names = NULL) {
 #'   (the default) the name is taken from the call to \code{x}.
 #' @param overwrite Logical; Indicates if already stored fit
 #'   indices should be overwritten. Defaults to \code{FALSE}.
+#'   Setting it to \code{TRUE} is useful for example when changing
+#'   additional arguments of an already stored criterion.
 #' @param file Either \code{NULL} or a character string. In the latter case, the
 #'   fitted model object including the newly added criterion values is saved via
 #'   \code{\link{saveRDS}} in a file named after the string supplied in
@@ -445,7 +492,9 @@ loo_model_weights.brmsfit <- function(x, ..., model_names = NULL) {
 #'   ignored otherwise. If \code{TRUE}, the fitted model object will be saved
 #'   regardless of whether new criteria were added via \code{add_criterion}.
 #' @param ... Further arguments passed to the underlying
-#'   functions computing the model fit criteria.
+#'   functions computing the model fit criteria. If you are recomputing
+#'   an already stored criterion with other \code{...} arguments, make
+#'   sure to set \code{overwrite = TRUE}.
 #'
 #' @return An object of the same class as \code{x}, but
 #'   with model fit criteria added for later usage.
@@ -623,25 +672,38 @@ validate_models <- function(models, model_names, sub_names) {
 
 # recommend options if approximate loo fails for some observations
 # @param moment_match has moment matching already been performed?
-recommend_loo_options <- function(loo, k_threshold, moment_match = FALSE,
+recommend_loo_options <- function(loo, k_threshold = 0.7, moment_match = FALSE,
                                   model_name = "") {
   if (isTRUE(nzchar(model_name))) {
     model_name <- paste0(" in model '", model_name, "'")
   } else {
     model_name <- ""
   }
-  n <- length(loo::pareto_k_ids(loo, threshold = k_threshold))
-  if (!moment_match && n > 0) {
+  ndraws <- dim(loo)[1] %||% Inf
+  n <- n2 <- length(loo::pareto_k_ids(loo, threshold = k_threshold))
+  # for small number of draws the threshold may be smaller than 0.7
+  k_threshold2 <- ps_khat_threshold(ndraws)
+  if (k_threshold2 < k_threshold) {
+    n2 <- length(loo::pareto_k_ids(loo, threshold = k_threshold2))
+  }
+  if (n2 > n && k_threshold2 <= 0.7) {
+    warning2(
+      "Found ", n2, " observations with a pareto_k > ", round(k_threshold2, 2),
+      model_name, ". We recommend to run more iterations to get at least ",
+      "about 2200 posterior draws to improve LOO-CV approximation accuracy."
+    )
+    out <- "loo_more_draws"
+  } else if (n > 0 && !moment_match) {
     warning2(
       "Found ", n, " observations with a pareto_k > ", k_threshold,
-      model_name, ". It is recommended to set 'moment_match = TRUE' in order ",
+      model_name, ". We recommend to set 'moment_match = TRUE' in order ",
       "to perform moment matching for problematic observations. "
     )
     out <- "loo_moment_match"
   } else if (n > 0 && n <= 10) {
     warning2(
       "Found ", n, " observations with a pareto_k > ", k_threshold,
-      model_name, ". It is recommended to set 'reloo = TRUE' in order to ",
+      model_name, ". We recommend to set 'reloo = TRUE' in order to ",
       "calculate the ELPD without the assumption that these observations " ,
       "are negligible. This will refit the model ", n, " times to compute ",
       "the ELPDs for the problematic observations directly."
@@ -661,6 +723,24 @@ recommend_loo_options <- function(loo, k_threshold, moment_match = FALSE,
   invisible(out)
 }
 
+# subset observations in a psis object
+# this is a bit cumbersome because of how psis stores information
+# @param subset vector with which to subset
+#' @export
+subset.psis <- function(x, subset, ...) {
+  stopifnot(is.vector(subset))
+  x$log_weights <- x$log_weights[, subset, drop = FALSE]
+  for (d in names(x$diagnostics)) {
+    x$diagnostics[[d]] <- x$diagnostics[[d]][subset]
+  }
+  attr_names <- c("norm_const_log", "tail_len", "r_eff")
+  for (a in attr_names) {
+    attr(x, a) <- attr(x, a)[subset]
+  }
+  attr(x, "dims") <- dim(x$log_weights)
+  x
+}
+
 # helper function to compute relative efficiences
 # @param x matrix of posterior draws
 # @param fit a brmsfit object to extract metadata from
@@ -936,3 +1016,9 @@ print.iclist <- function(x, digits = 2, ...) {
   print(round(mat, digits = digits), na.print = "")
   invisible(x)
 }
+
+# Pareto-smoothing k-hat threshold
+# not yet exported by loo so copied over here for now
+ps_khat_threshold <- function(S, ...) {
+  1 - 1 / log10(S)
+}
