@@ -53,7 +53,7 @@ extract <- function(x, ..., drop = FALSE, drop_dim = NULL) {
   }
   keep <- dim(out) > 1L | !drop_dim
   new_dim <- dim(out)[keep]
-  if (length(new_dim) == 1L) {
+  if (length(new_dim) <= 1L) {
     # use vectors instead of 1D arrays
     new_dim <- NULL
   }
@@ -135,7 +135,7 @@ find_elements <- function(x, ..., ls = list(), fun = '%in%') {
     stop("Argument 'ls' must be named.")
   }
   for (name in names(ls)) {
-    tmp <- lapply(x, "[[", name)
+    tmp <- from_list(x, name)
     out <- out & do_call(fun, list(tmp, ls[[name]]))
   }
   out
@@ -167,6 +167,9 @@ subset2 <- function(x, ..., ls = list(), fun = '%in%') {
   x[find_rows(x, ..., ls = ls, fun = fun), , drop = FALSE]
 }
 
+# not-in operator
+"%notin%" <- Negate("%in%")
+
 # convert array to list of elements with reduced dimension
 # @param x an arrary of dimension d
 # @return a list of arrays of dimension d-1
@@ -241,6 +244,10 @@ first_not_null <- function(...) {
   out
 }
 
+is_atomic_or_null <- function(x) {
+  is.atomic(x) || is.null(x)
+}
+
 isNA <- function(x) {
   length(x) == 1L && is.na(x)
 }
@@ -249,6 +256,15 @@ is_equal <- function(x, y, check.attributes = FALSE, ...) {
   isTRUE(all.equal(x, y, check.attributes = check.attributes, ...))
 }
 
+# extract factor levels from an arbitrary variable
+extract_levels <- function(x) {
+  # do not check for NAs according to #1355
+  if (!is.factor(x)) {
+    x <- factor(x)
+  }
+  levels(x)
+}
+
 # check if 'x' will behave like a factor in design matrices
 is_like_factor <- function(x) {
   is.factor(x) || is.character(x) || is.logical(x)
@@ -269,7 +285,7 @@ as_one_logical <- function(x, allow_na = FALSE) {
   s <- substitute(x)
   x <- as.logical(x)
   if (length(x) != 1L || anyNA(x) && !allow_na) {
-    s <- deparse_combine(s, max_char = 100L)
+    s <- deparse0(s, max_char = 100L)
     stop2("Cannot coerce '", s, "' to a single logical value.")
   }
   x
@@ -280,7 +296,7 @@ as_one_integer <- function(x, allow_na = FALSE) {
   s <- substitute(x)
   x <- SW(as.integer(x))
   if (length(x) != 1L || anyNA(x) && !allow_na) {
-    s <- deparse_combine(s, max_char = 100L)
+    s <- deparse0(s, max_char = 100L)
     stop2("Cannot coerce '", s, "' to a single integer value.")
   }
   x
@@ -291,7 +307,7 @@ as_one_numeric <- function(x, allow_na = FALSE) {
   s <- substitute(x)
   x <- SW(as.numeric(x))
   if (length(x) != 1L || anyNA(x) && !allow_na) {
-    s <- deparse_combine(s, max_char = 100L)
+    s <- deparse0(s, max_char = 100L)
     stop2("Cannot coerce '", s, "' to a single numeric value.")
   }
   x
@@ -302,7 +318,7 @@ as_one_character <- function(x, allow_na = FALSE) {
   s <- substitute(x)
   x <- as.character(x)
   if (length(x) != 1L || anyNA(x) && !allow_na) {
-    s <- deparse_combine(s, max_char = 100L)
+    s <- deparse0(s, max_char = 100L)
     stop2("Cannot coerce '", s, "' to a single character value.")
   }
   x
@@ -416,10 +432,35 @@ plapply <- function(X, FUN, cores = 1, ...) {
       on.exit(parallel::stopCluster(cl))
       out <- parallel::parLapply(cl = cl, X = X, fun = FUN, ...)
     }
+    # The version below hopefully prevents the spawning of zombies
+    # but it does not always succeed in that. It also seems to cause
+    # other issues as discussed in #1658, so commented out for now.
+    # cl_type <- ifelse(os_is_windows(), "PSOCK", "FORK")
+    # cl <- parallel::makeCluster(cores, type = cl_type)
+    # # Register a cleanup for the cluster in case the function fails
+    # # Need to wrap in a tryCatch to avoid error if cluster is already stopped
+    # on.exit(tryCatch(
+    #   { parallel::stopCluster(cl) },
+    #   error = function(e) invisible(NULL)
+    # ))
+    # out <- parallel::parLapply(cl = cl, X = X, fun = FUN, ...)
+    # parallel::stopCluster(cl)
   }
   out
 }
 
+# extract objects stored in each element of a list
+# @param x a list-like object
+# @param name name of the object to extract
+from_list <- function(x, name, ...) {
+  lapply(x, "[[", name, ...)
+}
+
+# unlist from_list output
+ufrom_list <- function(x, name, ..., recursive = TRUE, use.names = TRUE) {
+  unlist(from_list(x, name, ...), recursive, use.names)
+}
+
 # check if the operating system is Windows
 os_is_windows <- function() {
   isTRUE(Sys.info()[['sysname']] == "Windows")
@@ -574,7 +615,8 @@ rename <- function(x, pattern = NULL, replacement = NULL,
   dup <- duplicated(out)
   if (check_dup && any(dup)) {
     dup <- x[out %in% out[dup]]
-    stop2("Internal renaming led to duplicated names. \n",
+    stop2("Internal renaming led to duplicated names. ",
+          "Consider renaming your variables to have different suffixes.\n",
           "Occured for: ", collapse_comma(dup))
   }
   out
@@ -697,7 +739,7 @@ empty_data_frame <- function() {
 # @param value another named list-like object
 # @param dont_replace names of elements that cannot be replaced
 'replace_args<-' <- function(x, dont_replace = NULL, value) {
-  value_name <- deparse_combine(substitute(value), max_char = 100L)
+  value_name <- deparse0(substitute(value), max_char = 100L)
   value <- as.list(value)
   if (length(value) && is.null(names(value))) {
     stop2("Argument '", value_name, "' must be named.")
@@ -711,17 +753,18 @@ empty_data_frame <- function() {
   x
 }
 
-# deparse 'x' if it is no string
+# deparse0 'x' if it is no string
 deparse_no_string <- function(x) {
   if (!is.character(x)) {
-    x <- deparse_combine(x)
+    x <- deparse0(x)
   }
   x
 }
 
 # combine deparse lines into one string
-deparse_combine <- function(x, max_char = NULL) {
-  out <- collapse(deparse(x))
+# since R 4.0 we also have base::deparse1 for this purpose
+deparse0 <- function(x, max_char = NULL, ...) {
+  out <- collapse(deparse(x, ...))
   if (isTRUE(max_char > 0)) {
     out <- substr(out, 1L, max_char)
   }
@@ -754,7 +797,7 @@ eval_silent <- function(expr, type = "output", try = FALSE,
       try_out <- try(utils::capture.output(
         out <- eval(expr, envir), type = type, ...
       ))
-      if (is(try_out, "try-error")) {
+      if (is_try_error(try_out)) {
         # try again without suppressing error messages
         out <- eval(expr, envir)
       }
@@ -858,15 +901,15 @@ get_matches_expr <- function(pattern, expr, ...) {
   out <- NULL
   for (i in seq_along(expr)) {
     sexpr <- try(expr[[i]], silent = TRUE)
-    if (!is(sexpr, "try-error")) {
-      sexpr_char <- deparse_combine(sexpr)
+    if (!is_try_error(sexpr)) {
+      sexpr_char <- deparse0(sexpr)
       out <- c(out, get_matches(pattern, sexpr_char, ...))
     }
     if (is.call(sexpr) || is.expression(sexpr)) {
       out <- c(out, get_matches_expr(pattern, sexpr, ...))
     }
   }
-  unique(out)
+  trim_wsp(unique(out))
 }
 
 # like 'grepl' but handles (parse trees of) expressions
@@ -1012,6 +1055,11 @@ warn_deprecated <- function(new, old = as.character(sys.call(sys.parent()))[1])
   invisible(NULL)
 }
 
+# check if x is a try-error resulting from try()
+is_try_error <- function(x) {
+  inherits(x, "try-error")
+}
+
 # check if verbose mode is activated
 is_verbose <- function() {
   as_one_logical(getOption("brms.verbose", FALSE))
@@ -1025,69 +1073,6 @@ expect_match2 <- function(object, regexp, ..., all = TRUE) {
   testthat::expect_match(object, regexp, fixed = TRUE, ..., all = all)
 }
 
-# Copied from package 'vctrs' (more precisely:
-# <https://github.com/r-lib/vctrs/blob/master/R/register-s3.R>, version
-# 0.3.8.9001; identical to the code from version 0.3.8), as offered on the help
-# page for vctrs::s3_register() (version 0.3.8):
-s3_register_cp <- function(generic, class, method = NULL) {
-  stopifnot(is.character(generic), length(generic) == 1)
-  stopifnot(is.character(class), length(class) == 1)
-
-  pieces <- strsplit(generic, "::")[[1]]
-  stopifnot(length(pieces) == 2)
-  package <- pieces[[1]]
-  generic <- pieces[[2]]
-
-  caller <- parent.frame()
-
-  get_method_env <- function() {
-    top <- topenv(caller)
-    if (isNamespace(top)) {
-      asNamespace(environmentName(top))
-    } else {
-      caller
-    }
-  }
-  get_method <- function(method, env) {
-    if (is.null(method)) {
-      get(paste0(generic, ".", class), envir = get_method_env())
-    } else {
-      method
-    }
-  }
-
-  register <- function(...) {
-    envir <- asNamespace(package)
-
-    # Refresh the method each time, it might have been updated by
-    # `devtools::load_all()`
-    method_fn <- get_method(method)
-    stopifnot(is.function(method_fn))
-
-
-    # Only register if generic can be accessed
-    if (exists(generic, envir)) {
-      registerS3method(generic, class, method_fn, envir = envir)
-    } else if (identical(Sys.getenv("NOT_CRAN"), "true")) {
-      warning(sprintf(
-        "Can't find generic `%s` in package %s to register S3 method.",
-        generic,
-        package
-      ))
-    }
-  }
-
-  # Always register hook in case package is later unloaded & reloaded
-  setHook(packageEvent(package, "onLoad"), register)
-
-  # Avoid registration failures during loading (pkgload or regular)
-  if (isNamespaceLoaded(package)) {
-    register()
-  }
-
-  invisible()
-}
-
 # startup messages for brms
 .onAttach <- function(libname, pkgname) {
   version <- utils::packageVersion("brms")
@@ -1109,10 +1094,5 @@ s3_register_cp <- function(generic, class, method = NULL) {
       utils::packageVersion("emmeans") >= "1.4.0") {
     emmeans::.emm_register("brmsfit", pkgname)
   }
-  # dynamically register the 'get_refmodel.brmsfit' method for the
-  # 'get_refmodel' generic from 'projpred', if that package is installed
-  if (requireNamespace("projpred", quietly = TRUE)) {
-    s3_register_cp("projpred::get_refmodel", "brmsfit")
-  }
   invisible(NULL)
 }
