@@ -8,7 +8,9 @@
 #' @inheritParams predict.brmsfit
 #' @param x An \R object of class \code{brmsfit} or \code{loo} depending
 #'   on the method.
-#' @param loo An \R object of class \code{loo}.
+#' @param loo An \R object of class \code{loo}. If \code{NULL},
+#'   brms will try to extract a precomputed \code{loo} object
+#'   from the fitted model, added there via \code{\link{add_criterion}}.
 #' @param fit An \R object of class \code{brmsfit}.
 #' @param k_threshold The threshold at which Pareto \eqn{k}
 #'   estimates are treated as problematic. Defaults to \code{0.7}.
@@ -17,6 +19,9 @@
 #' @param check Logical; If \code{TRUE} (the default), some checks
 #'   check are performed if the \code{loo} object was generated
 #'   from the \code{brmsfit} object passed to argument \code{fit}.
+#' @param recompile Logical, indicating whether the Stan model should be
+#'   recompiled. This may be necessary if you are running \code{reloo} on
+#'   another machine than the one used to fit the model.
 #' @param future_args A list of further arguments passed to
 #'   \code{\link[future:future]{future}} for additional control over parallel
 #'   execution if activated.
@@ -44,20 +49,29 @@
 #' \dontrun{
 #' fit1 <- brm(count ~ zAge + zBase * Trt + (1|patient),
 #'             data = epilepsy, family = poisson())
+#'
 #' # throws warning about some pareto k estimates being too high
 #' (loo1 <- loo(fit1))
+#'
+#' # no more warnings after reloo
 #' (reloo1 <- reloo(fit1, loo = loo1, chains = 1))
 #' }
 #'
 #' @export
-reloo.brmsfit <- function(x, loo, k_threshold = 0.7, newdata = NULL,
-                          resp = NULL, check = TRUE, future_args = list(),
-                          ...) {
-  stopifnot(is.loo(loo), is.brmsfit(x), is.list(future_args))
+reloo.brmsfit <- function(x, loo = NULL, k_threshold = 0.7, newdata = NULL,
+                          resp = NULL, check = TRUE, recompile = NULL,
+                          future_args = list(), ...) {
+  stopifnot(is.brmsfit(x), is.list(future_args))
   if (is.brmsfit_multiple(x)) {
     warn_brmsfit_multiple(x)
     class(x) <- "brmsfit"
   }
+  loo <- loo %||% x$criteria[["loo"]]
+  if (is.null(loo)) {
+    stop2("No 'loo' object was provided and none is stored within the model.")
+  } else if (!is.loo(loo)) {
+    stop2("Inputs to the 'loo' argument must be of class 'loo'.")
+  }
   if (is.null(newdata)) {
     mf <- model.frame(x)
   } else {
@@ -91,24 +105,32 @@ reloo.brmsfit <- function(x, loo, k_threshold = 0.7, newdata = NULL,
     return(loo)
   }
 
+  # ensure that the model can be run in the current R session
+  x <- recompile_model(x, recompile = recompile)
+
   # split dots for use in log_lik and update
   dots <- list(...)
   ll_arg_names <- arg_names("log_lik")
   ll_arg_names <- intersect(names(dots), ll_arg_names)
   ll_args <- dots[ll_arg_names]
   ll_args$allow_new_levels <- TRUE
+  ll_args$sample_new_levels <-
+    first_not_null(ll_args$sample_new_levels, "gaussian")
   ll_args$resp <- resp
   ll_args$combine <- TRUE
   # cores is used in both log_lik and update
   up_arg_names <- setdiff(names(dots), setdiff(ll_arg_names, "cores"))
   up_args <- dots[up_arg_names]
+  up_args$object <- x
   up_args$refresh <- 0
 
   .reloo <- function(j) {
+    message(
+      "\nFitting model ", j, " out of ", J,
+      " (leaving out observation ", obs[j], ")"
+    )
     omitted <- obs[j]
     mf_omitted <- mf[-omitted, , drop = FALSE]
-    fit_j <- x
-    up_args$object <- fit_j
     up_args$newdata <- mf_omitted
     up_args$data2 <- subset_data2(x$data2, -omitted)
     fit_j <- SW(do_call(update, up_args))
@@ -118,25 +140,16 @@ reloo.brmsfit <- function(x, loo, k_threshold = 0.7, newdata = NULL,
     return(do_call(log_lik, ll_args))
   }
 
-  lls <- futures <- vector("list", J)
   message(
     J, " problematic observation(s) found.",
     "\nThe model will be refit ", J, " times."
   )
-  x <- recompile_model(x)
+  # TODO: separate parallel and non-parallel code to enable better printing?
+  future_args$X <- seq_len(J)
   future_args$FUN <- .reloo
-  future_args$seed <- TRUE
-  for (j in seq_len(J)) {
-    message(
-      "\nFitting model ", j, " out of ", J,
-      " (leaving out observation ", obs[j], ")"
-    )
-    future_args$args <- list(j)
-    futures[[j]] <- do_call("futureCall", future_args, pkg = "future")
-  }
-  for (j in seq_len(J)) {
-    lls[[j]] <- future::value(futures[[j]])
-  }
+  future_args$future.seed <- TRUE
+  lls <- do_call("future_lapply", future_args, pkg = "future.apply")
+
   # most of the following code is taken from rstanarm:::reloo
   # compute elpd_{loo,j} for each of the held out observations
   elpd_loo <- ulapply(lls, log_mean_exp)
