@@ -1,7 +1,7 @@
 params <-
 list(EVAL = TRUE)
 
-## ---- SETTINGS-knitr, include=FALSE-----------------------------------------------------
+## ----SETTINGS-knitr, include=FALSE------------------------------------------------------
 stopifnot(require(knitr))
 options(width = 90)
 opts_chunk$set(
@@ -9,8 +9,8 @@ opts_chunk$set(
   message = FALSE,
   warning = FALSE,
   eval = if (isTRUE(exists("params"))) params$EVAL else FALSE,
-  dev = "png",
-  dpi = 150,
+  dev = "jpeg",
+  dpi = 100,
   fig.asp = 0.8,
   fig.width = 5,
   out.width = "60%",
@@ -25,9 +25,10 @@ head(nhanes)
 
 ## ---------------------------------------------------------------------------------------
 library(mice)
-imp <- mice(nhanes, m = 5, print = FALSE)
+m <- 5
+imp <- mice(nhanes, m = m, print = FALSE)
 
-## ---- results = 'hide', message = FALSE-------------------------------------------------
+## ----results = 'hide', message = FALSE--------------------------------------------------
 fit_imp1 <- brm_multiple(bmi ~ age*chl, data = imp, chains = 2)
 
 ## ---------------------------------------------------------------------------------------
@@ -37,12 +38,19 @@ summary(fit_imp1)
 plot(fit_imp1, variable = "^b", regex = TRUE)
 
 ## ---------------------------------------------------------------------------------------
-round(fit_imp1$rhats, 2)
+library(posterior)
+draws <- as_draws_array(fit_imp1)
+# every dataset has nc = 2 chains in this example
+nc <- nchains(fit_imp1) / m
+draws_per_dat <- lapply(1:m, 
+  \(i) subset_draws(draws, chain = ((i-1)*nc+1):(i*nc))
+)
+lapply(draws_per_dat, summarise_draws, default_convergence_measures())
 
 ## ---------------------------------------------------------------------------------------
 conditional_effects(fit_imp1, "age:chl")
 
-## ---- results = 'hide', message = FALSE-------------------------------------------------
+## ----results = 'hide', message = FALSE--------------------------------------------------
 bform <- bf(bmi | mi() ~ age * mi(chl)) +
   bf(chl | mi() ~ age) + set_rescor(FALSE)
 fit_imp2 <- brm(bform, data = nhanes)
@@ -54,7 +62,7 @@ conditional_effects(fit_imp2, "age:chl", resp = "bmi")
 ## ---------------------------------------------------------------------------------------
 nhanes$se <- rexp(nrow(nhanes), 2)
 
-## ---- results = 'hide', message = FALSE, eval = FALSE-----------------------------------
+## ----results = 'hide', message = FALSE, eval = FALSE------------------------------------
 #  bform <- bf(bmi | mi() ~ age * mi(chl)) +
 #    bf(chl | mi(se) ~ age) + set_rescor(FALSE)
 #  fit_imp3 <- brm(bform, data = nhanes)
