@@ -34,29 +34,29 @@ test_that("restructure_formula_v1 works correctly", {
   expect_true(attr(form$formula, "nl"))
 })
 
-test_that("change_prior returns expected lists", {
+test_that("rename_prior returns expected lists", {
   pars <- c("b", "b_1", "bp", "bp_1", "prior_b", "prior_b__1",
             "prior_b__3", "sd_x[1]", "prior_bp__1")
   expect_equivalent(
-    brms:::change_prior(
+    brms:::rename_prior(
       class = "b", pars = pars, names = c("x1", "x3", "x2")
     ),
     list(list(pos = 6, fnames = "prior_b_x1"),
          list(pos = 7, fnames = "prior_b_x2"))
   )
   expect_equivalent(
-    brms:::change_prior(
+    brms:::rename_prior(
       class = "bp", pars = pars,
       names = c("x1", "x2"), new_class = "b"
     ),
     list(list(pos = 9, fnames = "prior_b_x1")))
 })
 
-test_that("change_old_re and change_old_re2 return expected lists", {
+test_that("rename_old_re and rename_old_re2 return expected lists", {
   data <- data.frame(y = rnorm(10), x = rnorm(10), g = 1:10)
   bterms <- brmsterms(bf(y ~ a, a ~ x + (1+x|g),
                         family = gaussian(), nl = TRUE))
-  ranef <- brms:::tidy_ranef(bterms, data = data)
+  ranef <- brms:::frame_re(bterms, data = data)
   target <- list(
     list(pos = c(rep(FALSE, 2), TRUE, rep(FALSE, 22)),
          oldname = "sd_a_g_Intercept", pnames = "sd_g_a_Intercept",
@@ -78,7 +78,7 @@ test_that("change_old_re and change_old_re2 return expected lists", {
             paste0("r_a_g[", 1:10, ",x]"))
   dims <- list("sd_a_g_Intercept" = numeric(0), "sd_a_g_x" = numeric(0),
                "cor_a_g_Intercept_x" = numeric(0), "r_a_g" = c(10, 2))
-  expect_equivalent(brms:::change_old_re(ranef, pars = pars, dims = dims), target)
+  expect_equivalent(brms:::rename_old_re(ranef, pars = pars, dims = dims), target)
 
   target <- list(
     list(pos = c(rep(FALSE, 2), TRUE, rep(FALSE, 22)),
@@ -101,10 +101,10 @@ test_that("change_old_re and change_old_re2 return expected lists", {
             paste0("r_g_a[", 1:10, ",x]"))
   dims <- list("sd_g_a_Intercept" = numeric(0), "sd_g_a_x" = numeric(0),
                "cor_g_a_Intercept_a_x" = numeric(0), "r_g_a" = c(10, 2))
-  expect_equivalent(brms:::change_old_re2(ranef, pars = pars, dims = dims), target)
+  expect_equivalent(brms:::rename_old_re2(ranef, pars = pars, dims = dims), target)
 })
 
-test_that("change_old_sm return expected lists", {
+test_that("rename_old_sm return expected lists", {
   target <- list(
     list(pos = c(FALSE, TRUE, rep(FALSE, 15)),
          oldname = "sds_sx1kEQ9",
@@ -134,10 +134,10 @@ test_that("change_old_sm return expected lists", {
                s_sigma_t2x0 = 6, s_sx1kEQ9 = 9)
   bterms <- brmsterms(bf(y ~ s(x1, k = 9), sigma ~ t2(x0)), family = gaussian())
   dat <- data.frame(y = rnorm(100), x1 = rnorm(100), x0 = rnorm(100))
-  expect_equivalent(brms:::change_old_sm(bterms, dat, pars, dims), target)
+  expect_equivalent(brms:::rename_old_sm(bterms, dat, pars, dims), target)
 })
 
-test_that("change_old_mo returns expected lists", {
+test_that("rename_old_mo returns expected lists", {
   bterms <- brmsterms(bf(y ~ mo(x), sigma ~ mo(x)), family = gaussian())
   data <- data.frame(y = rnorm(10), x = rep(1:5, 2))
   pars <- c(
@@ -163,10 +163,10 @@ test_that("change_old_mo returns expected lists", {
       fnames = paste0("simo_sigma_mox1[", 1:5, "]")
     )
   )
-  expect_equivalent(brms:::change_old_mo(bterms, data, pars), target)
+  expect_equivalent(brms:::rename_old_mo(bterms, data, pars), target)
 })
 
-test_that("change_old_categorical works correctly", {
+test_that("rename_old_categorical works correctly", {
   dat <- data.frame(
     y = rep(c("cat1", "cat2", "cat3"), 3),
     x = rnorm(9)
@@ -176,7 +176,7 @@ test_that("change_old_categorical works correctly", {
   bterms <- brmsterms(bf(y ~ x) + fam)
   pars <- c("b_cat2_Intercept", "b_cat3_Intercept",
             "b_cat2_x", "b_cat3_x")
-  res <- brms:::change_old_categorical(bterms, dat, pars)
+  res <- brms:::rename_old_categorical(bterms, dat, pars)
   target <- list(
     list(
       pos = rep(TRUE, 4),
